"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from collections import defaultdict
from sageopt.coniclifts.constraints.elementwise import ElementwiseConstraint
from sageopt.coniclifts.constraints.set_membership.psd_cone import PSD
from sageopt.coniclifts.utilities import array_index_iterator, __REAL_TYPES__


class ScalarAtom(object):

    # A ScalarAtom is a thing that is not reduced in a ScalarExpression.

    def is_convex(self):
        return False

    def is_concave(self):
        return False

    def is_affine(self):
        return False

    def __add__(self, other):
        return ScalarExpression({self: 1}, 0) + other

    def __sub__(self, other):
        return ScalarExpression({self: 1}, 0) - other

    def __mul__(self, other):
        if not isinstance(other, __REAL_TYPES__):
            raise RuntimeError('Can only multiply by scalars.')
        return ScalarExpression({self: other}, 0)

    def __truediv__(self, other):
        if not isinstance(other, __REAL_TYPES__):
            raise RuntimeError('Can only divide by scalars.')
        return ScalarExpression({self: (1 / other)}, 0)

    def __neg__(self):
        return ScalarExpression({self: -1}, 0)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rsub__(self, other):
        return other + ScalarExpression({self: -1}, 0)

    def __hash__(self):
        raise NotImplementedError()

    def __eq__(self, other):
        return NotImplementedError()

    def value(self):
        raise NotImplementedError()

    def scalar_variables(self):
        raise NotImplementedError()


class ScalarVariable(ScalarAtom):

    _SCALAR_VARIABLE_COUNTER = 0

    @staticmethod
    def curr_variable_count():
        return ScalarVariable._SCALAR_VARIABLE_COUNTER

    def __init__(self, parent, name):
        """
        :param parent: a Variable object originally containing this ScalarVariable
        :param name: A string; likely the parent's name followed by a subscript.
        """
        self._id = ScalarVariable._SCALAR_VARIABLE_COUNTER
        self._generation = parent.generation
        self._value = np.NaN
        self.name = name
        self.parent = parent
        ScalarVariable._SCALAR_VARIABLE_COUNTER += 1

    @property
    def id(self):
        return self._id

    def __hash__(self):
        return hash((self._id, self._generation))

    def __eq__(self, other):
        if isinstance(other, ScalarVariable):
            id_okay = self._id == other._id
            gen_okay = self._generation == other._generation
            return id_okay and gen_okay
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, ScalarVariable):
            return self._id < other._id
        else:
            raise RuntimeError('Cannot compare ScalarVariable to ' + str(type(other)))

    def __gt__(self, other):
        if isinstance(other, ScalarVariable):
            return self._id > other._id
        else:
            raise RuntimeError('Cannot compare ScalarVariable to ' + str(type(other)))

    def is_affine(self):
        return True

    def is_convex(self):
        return True

    def is_concave(self):
        return True

    def scalar_variables(self):
        return [self]

    def value(self):
        return self._value

    def __getstate__(self):
        # We lose the link to our parent.
        # Our parent will have to restore the link later.
        d = self.__dict__.copy()
        d['parent'] = None
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        pass


class NonlinearScalarAtom(ScalarAtom):

    @staticmethod
    def parse_arg(arg):
        # Type conversion
        if isinstance(arg, ScalarVariable):
            arg = ScalarExpression({arg: 1}, 0)
        elif isinstance(arg, Expression) and arg.size == 1:
            arg = np.asscalar(arg)
            # noinspection PyTypeChecker
            return NonlinearScalarAtom.parse_arg(arg)
        elif isinstance(arg, __REAL_TYPES__):
            arg = ScalarExpression(dict(), arg)
        # Parsing
        if isinstance(arg, ScalarExpression) and arg.is_affine():
            arg.remove_zeros()
            res = sorted(list(arg.atoms_to_coeffs.items()))
            return tuple(res + [('OFFSET', arg.offset)])
        else:
            raise RuntimeError('ScalarAtom arguments must be affine ScalarExpressions.')

    @staticmethod
    def __atom_text__():
        raise NotImplementedError()

    @property
    def id(self):
        # noinspection PyUnresolvedReferences
        return self._id

    @property
    def args(self):
        # noinspection PyUnresolvedReferences
        return self._args

    @property
    def evaluator(self):
        # noinspection PyUnresolvedReferences
        return self._evaluator

    @property
    def epigraph_variable(self):
        # noinspection PyUnresolvedReferences
        return self._epigraph_variable

    def __hash__(self):
        return hash(self.args + (self.__atom_text__(),))

    def __eq__(self, other):
        if not type(self) == type(other):
            return False
        else:
            return self.args == other.args

    def scalar_variables(self):
        var_list = []
        for idx in range(len(self.args)):
            var_list += [tup[0] for tup in self.args[idx][:-1]]
        return var_list

    def value(self):
        f = self.evaluator
        # construct numeric values for argument expressions
        args_val = np.zeros(shape=(len(self.args),))
        for idx, arg in enumerate(self.args):
            var_vals = np.array([tup[0].value() for tup in arg[:-1]])
            var_coeffs = np.array([tup[1] for tup in arg[:-1]])
            args_val[idx] = np.dot(var_coeffs, var_vals) + arg[-1][1]
        # evaluate the defining function, and return the result
        val = f(args_val)
        return val

    def is_convex(self):
        raise NotImplementedError()

    def is_concave(self):
        raise NotImplementedError()

    def epigraph_conic_form(self):
        """
        :return: A_vals, A_rows, A_cols, b, K, _epigraph_variable, sep_K
            A_vals - list (of floats)
            A_rows - numpy 1darray (of integers)
            A_cols - list (of integers)
            b - numpy 1darray (of floats)
            K - list (of coniclifts Cone objects)
        """
        raise NotImplementedError()


class ScalarExpression(object):

    __array_priority__ = 100

    def __init__(self, atoms_to_coeffs, offset, verify=True):
        """

        :param atoms_to_coeffs: a dictionary mapping ScalarAtoms to numeric types.
        :param offset: a numeric type.

        Represents sum([ c * v for (v,c) in atoms_to_coeffs.items()]) + offset.
        """
        if not isinstance(atoms_to_coeffs, defaultdict):
            self.atoms_to_coeffs = defaultdict(int)
            if verify:
                if not isinstance(offset, __REAL_TYPES__):
                    raise RuntimeError('Coefficients in ScalarExpressions can only be numeric types.')
                if not all(isinstance(v, __REAL_TYPES__) for v in atoms_to_coeffs.values()):
                    raise RuntimeError('Coefficients in ScalarExpressions can only be numeric types.')
                if not all(isinstance(v, ScalarAtom) for v in atoms_to_coeffs.keys()):
                    raise RuntimeError('Keys in ScalarExpressions must be ScalarAtoms.')
            self.atoms_to_coeffs.update(atoms_to_coeffs)
        else:
            self.atoms_to_coeffs = atoms_to_coeffs.copy()
        self.offset = offset

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            return self.as_expr() + other
        f = ScalarExpression(self.atoms_to_coeffs, self.offset, verify=False)
        if isinstance(other, __REAL_TYPES__):
            f.offset += other
        else:
            if isinstance(other, ScalarAtom):
                other = ScalarExpression({other: 1}, 0)
            for k, v in other.atoms_to_coeffs.items():
                f.atoms_to_coeffs[k] += v
            f.offset += other.offset
        return f

    def __sub__(self, other):
        if isinstance(other, np.ndarray):
            return self.as_expr() - other
        if isinstance(other, __REAL_TYPES__):
            f = ScalarExpression(self.atoms_to_coeffs, self.offset, verify=False)
            f.offset -= other
        else:
            if isinstance(other, ScalarAtom):
                other = ScalarExpression({other: 1}, 0)
            f = ScalarExpression(self.atoms_to_coeffs, self.offset, verify=False)
            for k, v in other.atoms_to_coeffs.items():
                f.atoms_to_coeffs[k] -= v
            f.offset -= other.offset
        return f

    def __mul__(self, other):
        if isinstance(other, __REAL_TYPES__):
            if other == 0:
                return ScalarExpression(dict(), 0)
            f = ScalarExpression(self.atoms_to_coeffs, self.offset, verify=False)
            for k in f.atoms_to_coeffs:
                f.atoms_to_coeffs[k] *= other
            f.offset *= other
            return f
        elif isinstance(other, np.ndarray):
            return self.as_expr() * other
        elif isinstance(other, ScalarExpression) and other.is_constant():
            return other.offset * self
        else:
            return other.__rmul__(self)

    def __truediv__(self, other):
        if not isinstance(other, __REAL_TYPES__):
            print(type(other))
            raise RuntimeError('Can only divide ScalarExpressions by scalars.')
        return self * (1 / other)

    def __neg__(self):
        f = ScalarExpression(self.atoms_to_coeffs, self.offset, verify=False)
        for k in f.atoms_to_coeffs:
            f.atoms_to_coeffs[k] *= -1
        f.offset *= -1
        return f

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rsub__(self, other):
        # noinspection PyTypeChecker
        return other + (-1) * self

    def remove_zeros(self):
        acs = list(self.atoms_to_coeffs.items())
        for a, c in acs:
            if c == 0:
                del self.atoms_to_coeffs[a]

    def is_constant(self):
        return len(self.atoms_to_coeffs) == 0

    def is_convex(self):
        for a, c in self.atoms_to_coeffs.items():
            if a.is_affine():
                continue
            elif a.is_convex() and c < 0:
                return False
            elif a.is_concave() and c > 0:
                return False
            elif not (a.is_concave() or a.is_convex()):
                return False
        return True

    def is_concave(self):
        for a, c in self.atoms_to_coeffs.items():
            if a.is_affine():
                continue
            elif a.is_convex() and c > 0:
                return False
            elif a.is_concave() and c < 0:
                return False
            elif not (a.is_concave() or a.is_convex()):
                return False
        return True

    def is_affine(self):
        return all(isinstance(k, ScalarVariable) for k in self.atoms_to_coeffs)

    def __le__(self, other):
        # self <= other
        return ElementwiseConstraint(self, other, "<=")

    def __ge__(self, other):
        # self >= other
        return ElementwiseConstraint(self, other, ">=")

    def __eq__(self, other):
        return ElementwiseConstraint(self, other, "==")

    def scalar_variables(self):
        svs = []
        for a in self.atoms_to_coeffs:
            if isinstance(a, ScalarVariable):
                svs.append(a)
            else:
                svs += a.scalar_variables()
        return list(set(svs))

    def variables(self):
        var_ids = set()
        var_list = []
        sv = self.scalar_variables()
        for v in sv:
            if id(v.parent) not in var_ids:
                var_ids.add(id(v.parent))
                var_list.append(v.parent)
        return var_list

    def as_expr(self):
        # noinspection PyTypeChecker
        return np.array(self).view(Expression)

    def value(self):
        atoms_and_coeffs = list(self.atoms_to_coeffs.items())
        atom_vals = np.array([ac[0].value() for ac in atoms_and_coeffs])
        atom_coeffs = np.array([ac[1] for ac in atoms_and_coeffs])
        val = np.dot(atom_vals, atom_coeffs) + self.offset
        return val


class Expression(np.ndarray):

    """
    SUBCLASSING NDARRAY

    https://docs.scipy.org/doc/numpy/user/basics.subclassing.html#basics-subclassing
    """

    def __new__(cls, obj):
        attempt = np.array(obj, dtype=object, copy=False, subok=True)
        for tup in array_index_iterator(attempt.shape):
            # noinspection PyTypeChecker,PyCallByClass
            Expression.__setitem__(attempt, tup, attempt[tup])
        return attempt.view(Expression)

    def __matmul__(self, other):
        """
        :param other: "self" is multiplying "other" on the left.
        :return: self @ other
        """
        if other.ndim > 2 or self.ndim > 2:
            msg = '\n \t Matmul implementation uses "dot", '
            msg += 'which behaves differently for higher dimension arrays.\n'
            raise RuntimeError(msg)
        return Expression.__rmatmul__(self.T, other.T).T

    def __rmatmul__(self, other):
        """
        :param other: a constant Expression or nd-array which left-multiplies "self".
        :return: other @ self
        """
        if other.ndim > 2 or self.ndim > 2:
            msg = '\n \t Matmul implementation uses "dot", '
            msg += 'which behaves differently for higher dimension arrays.\n'
            raise RuntimeError(msg)
        (A, x, B) = self.factor()
        if isinstance(other, Expression):
            if not other.is_constant():
                raise RuntimeError('Can only multiply by constant Expressions.')
            else:
                _, _, other = other.factor()
        if other.ndim == 2:
            other_times_A = np.tensordot(other, A, axes=1)
        else:
            other_times_A = np.tensordot(other.reshape((1, -1)), A, axes=1)
            other_times_A = np.squeeze(other_times_A, axis=0)
        other_times_A_x = Expression.disjoint_dot(other_times_A, x)
        res = other_times_A_x
        other_times_B = np.tensordot(other, B, axes=1)
        for tup in array_index_iterator(other_times_A_x.shape):
            res[tup].offset = other_times_B[tup]
        return res

    def __le__(self, other):
        return ElementwiseConstraint(self, other, "<=")

    def __ge__(self, other):
        return ElementwiseConstraint(self, other, ">=")

    def __eq__(self, other):
        return ElementwiseConstraint(self, other, "==")

    def __lshift__(self, other):
        # self << other
        return PSD(other - self)

    def __rshift__(self, other):
        # self >> other
        return PSD(self - other)

    def __setitem__(self, key, value):
        if isinstance(value, ScalarExpression):
            np.ndarray.__setitem__(self, key, value)
        elif isinstance(value, __REAL_TYPES__):
            np.ndarray.__setitem__(self, key, ScalarExpression(dict(), value, verify=False))
        elif isinstance(value, ScalarAtom):
            np.ndarray.__setitem__(self, key, ScalarExpression({value: 1}, 0, verify=False))
        elif isinstance(value, np.ndarray) and value.size == 1:
            # noinspection PyTypeChecker
            self[key] = np.asscalar(value)
        elif isinstance(value, Expression):
            np.ndarray.__setitem__(self, key, value)
        else:
            raise RuntimeError('Can only initialize with numeric, ScalarExpression, or ScalarAtom types.')

    def scalar_atoms(self):
        return list(set.union(*[set(se.scalar_atoms()) for se in self.flat]))

    def scalar_variables(self):
        return list(set.union(*[set(se.scalar_variables()) for se in self.flat]))

    def variables(self):
        var_ids = set()
        var_list = []
        for se in self.ravel():
            for sv in se.scalar_variables():
                if id(sv.parent) not in var_ids:
                    var_ids.add(id(sv.parent))
                    var_list.append(sv.parent)
        return var_list

    def is_affine(self):
        if not Expression.can_assume_affine(self):
            return all(v.is_affine() for v in self.flat)
        else:
            return True

    def is_constant(self):
        return all(v.is_constant() for v in self.flat)

    def is_convex(self):
        res = np.empty(shape=self.shape, dtype=bool)
        for tup in array_index_iterator(self.shape):
            res[tup] = self[tup].is_convex()
        return res

    def is_concave(self):
        res = np.empty(shape=self.shape, dtype=bool)
        for tup in array_index_iterator(self.shape):
            res[tup] = self[tup].is_concave()
        return res

    def as_expr(self):
        return self

    def factor(self):
        """
        :return: (A, x, b) -- A is a numpy ndarray of numeric scalars. x is an Expression object
        where each entry in x.ravel() is a ScalarExpression consisting of a single atom
        (with zero offset), and b is a numpy array of same dimensions as the Expression
        np.matmul(A, x).
        """
        x = list(set(a for se in self.flat for a in se.atoms_to_coeffs))
        x.sort(key=lambda a: a.id)
        # Sorting by ScalarAtom id makes this method deterministic
        # when all ScalarAtoms in this Expression are of the same type.
        # That's useful for, e.g. affine Expressions, which
        # we need to test for symbolic equivalence.
        atoms_to_pos = {a: i for (i, a) in enumerate(x)}
        A = np.zeros(self.shape + (len(x),))
        B = np.zeros(self.shape)
        for tup in array_index_iterator(self.shape):
            se = self[tup]
            for a, c in se.atoms_to_coeffs.items():
                A[tup + (atoms_to_pos[a],)] = c
            B[tup] = se.offset
        return A, x, B

    @property
    def value(self):
        val = np.zeros(shape=self.shape)
        for tup in array_index_iterator(self.shape):
            val[tup] = self[tup].value()
        return val

    @staticmethod
    def disjoint_dot(array, list_of_atoms):
        # This is still MUCH SLOWER than adding numbers together.
        if len(list_of_atoms) != array.shape[-1]:
            raise RuntimeError('Incompatible dimensions to disjoint_dot.')
        expr = np.empty(shape=array.shape[:-1], dtype=object)
        for tup in array_index_iterator(expr.shape):
            dict_items = []
            for i, a in enumerate(list_of_atoms):
                dict_items.append((a, array[tup + (i,)]))
            d = dict(dict_items)
            expr[tup] = ScalarExpression(d, 0, verify=False)
        return expr.view(Expression)

    @staticmethod
    def disjoint_add(expr1, expr2):
        if expr1.shape != expr2.shape:
            raise RuntimeError('Disjoint add requires operands of the same shape.')
        expr3 = np.empty(expr1.shape, dtype=object)
        for tup in array_index_iterator(expr3.shape):
            d = dict(expr1[tup].atoms_to_coeffs)
            d.update(expr2[tup].atoms_to_coeffs)
            offset = expr1[tup].offset + expr2[tup].offset
            expr3[tup] = ScalarExpression(d, offset)
        return expr3.view(Expression)

    @staticmethod
    def can_assume_affine(expr):
        unadorned_var = isinstance(expr, (Variable, ScalarVariable))
        constant_array = isinstance(expr, np.ndarray) and expr.dtype != object
        numeric = isinstance(expr, __REAL_TYPES__)
        return unadorned_var or constant_array or numeric

    @staticmethod
    def are_equivalent(expr1, expr2, rtol=1e-5, atol=1e-8):
        # WARNING:
        #   The behavior of this function is conservative.
        #   Non-deterministic behavior of factor(...) when either Expression
        #   contains a mix of different ScalarAtoms (e.g. ScalarVariables
        #   and NonlinearScalarAtoms) can cause this function to return
        #   False when the expressions are in fact equivalent.
        #
        #   This function only works on 0d, 1d, and 2d expressions.
        if not isinstance(expr1, Expression):
            expr1 = Expression(expr1)
        if not isinstance(expr2, Expression):
            expr2 = Expression(expr2)
        if expr1.shape != expr2.shape:
            return False
        A1, x1, B1 = expr1.factor()
        A2, x2, B2 = expr2.factor()
        for i in range(len(x1)):
            if not isinstance(x2[i], type(x1[i])):
                return False
            elif x1[i].id != x2[i].id:
                return False
        if not np.allclose(A1, A2, rtol=rtol, atol=atol):
            return False
        elif not np.allclose(B1, B2, rtol=rtol, atol=atol):
            return False
        else:
            return True


class Variable(Expression):

    _UNNAMED_VARIABLE_CALL_COUNT = 0

    _VARIABLE_GENERATION = 0

    # noinspection PyInitNewSignature
    def __new__(cls, shape=(), name=None, var_properties=None):
        if var_properties is None:
            var_properties = []
        if name is None:
            name = 'unnamed_var_' + str(Variable._UNNAMED_VARIABLE_CALL_COUNT)
            Variable._UNNAMED_VARIABLE_CALL_COUNT += 1
        obj = np.empty(shape=shape, dtype=object).view(Variable)
        obj._is_proper = True
        obj._name = name
        obj._var_properties = var_properties
        obj._scalar_variable_ids = []
        obj._generation = Variable._VARIABLE_GENERATION
        if len(var_properties) == 0:
            Variable.__unstructured_populate__(obj, name)
        elif 'symmetric' in var_properties:
            Variable.__symmetric_populate__(obj, name)
        else:
            Variable.__unstructured_populate__(obj, name)
            raise UserWarning('The variable with name ' + name + ' was declared with an unknown property.')
        return obj

    # noinspection PyProtectedMember
    @staticmethod
    def __unstructured_populate__(obj, name):
        if obj.shape == ():
            v = ScalarVariable(parent=obj, name=name)
            np.ndarray.__setitem__(obj, tuple(), ScalarExpression({v: 1}, 0, verify=False))
            obj._scalar_variable_ids.append(v.id)
        else:
            for tup in array_index_iterator(obj.shape):
                v = ScalarVariable(parent=obj, name=name + str(list(tup)))
                obj._scalar_variable_ids.append(v.id)
                np.ndarray.__setitem__(obj, tup, ScalarExpression({v: 1}, 0, verify=False))
        pass

    # noinspection PyProtectedMember
    @staticmethod
    def __symmetric_populate__(obj, name):
        if obj.ndim != 2 or obj.shape[0] != obj.shape[1]:
            raise RuntimeError('Symmetric variables must be 2d, and square.')
        temp_id_array = np.zeros(shape=obj.shape, dtype=int)
        for i in range(obj.shape[0]):
            v = ScalarVariable(parent=obj, name=name + str([i, i]))
            np.ndarray.__setitem__(obj, (i, i), ScalarExpression({v: 1}, 0, verify=False))
            temp_id_array[i, i] = v.id
            for j in range(i+1, obj.shape[1]):
                v = ScalarVariable(parent=obj, name=name + str([i, j]))
                np.ndarray.__setitem__(obj, (i, j), ScalarExpression({v: 1}, 0, verify=False))
                np.ndarray.__setitem__(obj, (j, i), ScalarExpression({v: 1}, 0, verify=False))
                temp_id_array[i, j] = v.id
                temp_id_array[j, i] = v.id
        for tup in array_index_iterator(obj.shape):
            obj._scalar_variable_ids.append(temp_id_array[tup])
        pass

    def __setitem__(self, key, value):
        raise RuntimeError('Cannot reassign entries of Variable objects.')

    def __add__(self, other):
        return Expression.__add__(self.view(Expression), other)

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        return Expression.__mul__(self.view(Expression), other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return Expression.__neg__(self.view(Expression))

    def __sub__(self, other):
        return Expression.__sub__(self.view(Expression), other)

    def __rsub__(self, other):
        return other + self.__neg__()

    def __rmatmul__(self, other):
        return Expression.__rmatmul__(self.view(Expression), other)

    # noinspection PyMethodOverriding
    def __setstate__(self, state):
        np.ndarray.__setstate__(self, state)
        self.relink_scalar_variables()

    def is_constant(self):
        return False

    def is_affine(self):
        return True

    def scalar_variables(self):
        return [list(se.atoms_to_coeffs)[0] for se in self.flat]

    def leading_scalar_variable_id(self):
        if self.is_proper():
            return self._scalar_variable_ids[0]
        else:
            return self[(0,) * len(self.shape)].scalar_variables()[0].id

    def set_scalar_variables(self, value):
        if value.shape != self.shape:
            raise RuntimeError('Dimension mismatch.')
        for tup in array_index_iterator(self.shape):
            sv = list(self[tup].atoms_to_coeffs)[0]
            sv._value = value[tup]
        pass

    @property
    def scalar_variable_ids(self):
        if self.is_proper():
            return self._scalar_variable_ids
        else:
            return [self[tup].scalar_variables()[0].id for tup in array_index_iterator(self.shape)]

    @property
    def name(self):
        if hasattr(self, '_name'):
            return self._name
        else:
            tup = (0,) * len(self.shape)
            temp = self[(0,) * len(self.shape)].scalar_variables()[0].name
            if self.ndim > 0:
                # Need to trim the indices at the end of the string
                return temp[:-len(str(list(tup)))]
            else:
                return temp

    @property
    def generation(self):
        return self._generation

    def is_proper(self):
        return hasattr(self, '_is_proper')

    def relink_scalar_variables(self):
        if not self.is_proper():
            pass
        svs = self.scalar_variables()
        for sv in svs:
            sv.parent = self




"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt import coniclifts as cl
from sageopt.symbolic.polynomials import Polynomial
from sageopt.symbolic.signomials import Signomial
from sageopt.relaxations import sage_sigs
from sageopt.relaxations import constraint_generators as con_gen
from sageopt.relaxations import symbolic_correspondences as sym_corr


def primal_sage_poly_cone(poly, name, log_AbK):
    poly_sr, poly_sr_cons = poly.sig_rep
    con = sage_sigs.primal_sage_cone(poly_sr, name, log_AbK)
    constrs = [con] + poly_sr_cons
    return constrs


def relative_dual_sage_poly_cone(primal_poly, dual_var, name_base, log_AbK):
    """
    :param log_AbK:
    :param primal_poly: a Polynomial
    :param dual_var: a coniclifts Variable with y.shape == (p.m, 1).
    :param name_base:

    :return: coniclifts Constraints over y (and additional auxilliary variables, as
    necessary) so that y defines a dual variable to the constraint that "p is a SAGE polynomial."
    """
    sr, sr_cons = primal_poly.sig_rep
    evens = [i for i, row in enumerate(sr.alpha) if np.all(row % 2 == 0)]
    if len(evens) < sr.m:
        is_even = np.zeros(shape=(sr.m,), dtype=bool)
        is_even[evens] = True
        aux_v = cl.Variable(shape=(sr.m, 1), name='aux_v_{' + name_base + ' sage poly dual}')
        constrs = [sage_sigs.relative_dual_sage_cone(sr, aux_v, name_base + ' sigrep sage dual', log_AbK),
                   aux_v[is_even] == dual_var[is_even],
                   -aux_v[~is_even] <= dual_var[~is_even], dual_var[~is_even] <= aux_v[~is_even]]
    else:
        constrs = [sage_sigs.relative_dual_sage_cone(sr, dual_var, name_base + ' sigrep sage dual', log_AbK)]
    return constrs


def poly_relaxation(f, form='dual', poly_ell=0, sigrep_ell=0, X=None):
    """
    Construct a coniclifts Problem instance for producing a lower bound on

    .. math::

        \min\{ f(x) \,:\, x \\in X \}

    where :math:`X=R^{\\texttt{f.n}}` by default.

    If ``form='dual'``, we can also attempt to recover solutions to the above problem.

    Parameters
    ----------
    f : Polynomial
        The objective function to be minimized.
    form : str
        Either ``form='primal'`` or ``form='dual'``.
    poly_ell : int
        Controls the complexity of a polynomial modulating function. Must be nonnegative.
    sigrep_ell : int
        Controls the complexity of the SAGE certificate applied to the Lagrangian's signomial representative.
    X : dict
        If ``X`` is None, then we produce a bound on ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'log_AbK'``, ``'gts'``, and ``'eqs'``.
        For most situations, the appropriate dict ``X`` can be generated with ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    """
    form = form.lower()
    if form[0] == 'd':
        prob = poly_dual(f, poly_ell, sigrep_ell, X)
    elif form[0] == 'p':
        prob = poly_primal(f, poly_ell, sigrep_ell, X)
    else:
        raise RuntimeError('Unrecognized form: ' + form + '.')
    return prob


def poly_dual(f, poly_ell=0, sigrep_ell=0, X=None):
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    if poly_ell == 0:
        sr, cons = f.sig_rep
        if len(cons) > 0:
            msg = '\n\nThe provided Polynomial has nonconstant coefficients.\n'
            msg += 'The most likely cause is that a mistake has been made in setting up '
            msg += 'the data for this function.\n Raising a RuntimeError.\n'
            raise RuntimeError(msg)
        log_X = {'AbK':  X['log_AbK'],
                 'gts': [log_domain_converter(g) for g in X['gts']],
                 'eqs': [log_domain_converter(g) for g in X['eqs']]}
        prob = sage_sigs.sig_dual(sr, sigrep_ell, X=log_X)
        cl.clear_variable_indices()
        return prob
    elif sigrep_ell == 0:
        modulator = f.standard_multiplier() ** poly_ell
        gamma = cl.Variable()
        lagrangian = (f - gamma) * modulator
        v = cl.Variable(shape=(lagrangian.m, 1), name='v')
        con_base_name = v.name + ' domain'
        constraints = relative_dual_sage_poly_cone(lagrangian, v, con_base_name, log_AbK=X['log_AbK'])
        a = sym_corr.relative_coeff_vector(modulator, lagrangian.alpha)
        constraints.append(a.T @ v == 1)
        f_mod = Polynomial(f.alpha_c) * modulator
        obj_vec = sym_corr.relative_coeff_vector(f_mod, lagrangian.alpha)
        obj = obj_vec.T @ v
        prob = cl.Problem(cl.MIN, obj, constraints)
        cl.clear_variable_indices()
        return prob
    else:
        raise NotImplementedError()


def poly_primal(f, poly_ell=0, sigrep_ell=0, X=None):
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    if poly_ell == 0:
        sr, cons = f.sig_rep
        if len(cons) > 0:
            msg = '\n\nThe provided Polynomial has nonconstant coefficients.\n'
            msg += 'The most likely cause is that a mistake has been made in setting up '
            msg += 'the data for this function.\n Raising a RuntimeError.\n'
            raise RuntimeError(msg)
        log_X = {'AbK':  X['log_AbK'],
                 'gts': [log_domain_converter(g) for g in X['gts']],
                 'eqs': [log_domain_converter(g) for g in X['eqs']]}
        prob = sage_sigs.sig_primal(sr, sigrep_ell, X=log_X, additional_cons=cons)
        cl.clear_variable_indices()
        return prob
    else:
        poly_modulator = f.standard_multiplier() ** poly_ell
        gamma = cl.Variable(shape=(), name='gamma')
        lagrangian = (f - gamma) * poly_modulator
        if sigrep_ell > 0:
            sr, cons = lagrangian.sig_rep
            sig_modulator = Signomial(sr.alpha, np.ones(shape=(sr.m,))) ** sigrep_ell
            sig_under_test = sr * sig_modulator
            con_name = 'Lagrangian modulated sigrep sage'
            con = sage_sigs.primal_sage_cone(sig_under_test, con_name, AbK=X['log_AbK'])
            constraints = [con] + cons
        else:
            con_name = 'Lagrangian sage poly'
            constraints = primal_sage_poly_cone(lagrangian, con_name, log_AbK=X['log_AbK'])
        obj = gamma
        prob = cl.Problem(cl.MAX, obj, constraints)
        cl.clear_variable_indices()
        return prob


def sage_feasibility(f, X=None):
    """
    Constructs a coniclifts maximization Problem which is feasible iff ``f`` admits an X-SAGE decomposition.

    Parameters
    ----------
    f : Polynomial
        We want to test if this function admits an X-SAGE decomposition.
    X : dict
        If ``X`` is None, then we test nonnegativity of ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'log_AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
        A coniclifts maximization Problem. If ``f`` admits an X-SAGE decomposition, then we should have
        ``prob.value > -np.inf``, once ``prob.solve()`` has been called.
    """
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    log_X = {'AbK': X['log_AbK'],
             'gts': [log_domain_converter(g) for g in X['gts']],
             'eqs': [log_domain_converter(g) for g in X['eqs']]}
    sr, cons = f.sig_rep
    prob = sage_sigs.sage_feasibility(sr, X=log_X, additional_cons=cons)
    cl.clear_variable_indices()
    return prob


def sage_multiplier_search(f, level=1, X=None):
    """
    Constructs a coniclifts maximization Problem which is feasible if ``f`` can be certified as nonnegative
    over ``X``, by using an appropriate X-SAGE modulating function.

    Parameters
    ----------
    f : Polynomial
        We want to test if ``f`` is nonnegative over ``X``.
    level : int
        Controls the complexity of the X-SAGE modulating function. Must be a positive integer.
    X : dict
        If ``X`` is None, then we test nonnegativity of ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'log_AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    Notes
    -----
    This function provides an alternative to moving up the SAGE hierarchy, for the goal of certifying
    nonnegativity of a polynomial ``f`` over some set ``X`` where ``|X|`` is log-log convex.
    In general, the approach is to introduce a polynomial

        ``mult = Polynomial(alpha_hat, c_tilde)``

    where the rows of alpha_hat are all "level"-wise sums of rows from ``f.alpha``, and ``c_tilde``
    is a coniclifts Variable defining a nonzero SAGE polynomial. Then we can check if
    ``f_mod = f * mult`` is SAGE for any choice of ``c_tilde``.
    """
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    constraints = []
    # Make the multiplier polynomial (and require that it be SAGE)
    mult_alpha = hierarchy_e_k([f], k=level)
    c_tilde = cl.Variable(shape=(mult_alpha.shape[0],), name='c_tilde')
    mult = Polynomial(mult_alpha, c_tilde)
    temp_cons = primal_sage_poly_cone(mult, name=(c_tilde.name + ' domain'), log_AbK=X['log_AbK'])
    constraints += temp_cons
    constraints.append(cl.sum(c_tilde) >= 1)
    # Make "f_mod := f * mult", and require that it be SAGE.
    f_mod = mult * f
    temp_cons = primal_sage_poly_cone(f_mod, name='f_mod sage poly', log_AbK=X['log_AbK'])
    constraints += temp_cons
    # noinspection PyTypeChecker
    prob = cl.Problem(cl.MAX, 0, constraints)
    cl.clear_variable_indices()
    return prob


def poly_constrained_relaxation(f, gts, eqs, form='dual', p=0, q=1, ell=0, X=None):
    """
    Construct a coniclifts Problem instance representing a level-``(p, q, ell)`` SAGE relaxation
    for the polynomial optimization problem

    .. math::

        \\begin{align*}
          \min\{ f(x) :~& g(x) \geq 0 \\text{ for } g \\in \\text{gts}, \\\\
                       & g(x) = 0  \\text{ for } g \\in \\text{eqs}, \\\\
                       & \\text{and } x \\in X \}
        \\end{align*}

    where :math:`X = R^{\\texttt{f.n}}` by default. The optimal value of this relaxation will
    produce a lower bound on the minimization problem described above. When ``form='dual'``,
    a solution to this  relaxation can be used to help recover optimal solutions to the problem
    described above.

    Parameters
    ----------
    f : Polynomial
        The objective to be minimized.
    gts : list of Polynomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Polynomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    form : str
        Either ``form='primal'`` or ``form='dual'``.
    p : int
        Controls the complexity of Lagrange multipliers in the primal formulation, and (equivalently) constraints in
        the dual formulation. The smallest value is ``p=0``, which corresponds to scalar Lagrange multipliers.
    q : int
        The number of folds applied to the constraints ``gts`` and ``eqs``. The smallest value is ``q=1``, which
        means "leave ``gts`` and ``eqs`` as-is."
    ell : int
        Controls the complexity of any modulator applied to the Lagrangian in the primal formulation, and
        (equivalently) constraints in the dual formulation. The smallest value is ``ell=0``, which means
        the primal Lagrangian must be a SAGE polynomial.
    X : dict
        If not-None, then ``X`` must be dictionary with three keys ``'log_AbK'``, ``'gts'``, ``'eqs'`` such as that
        generated by the function ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
    """
    form = form.lower()
    if form[0] == 'd':
        prob = poly_constrained_dual(f, gts, eqs, p, q, ell, X)
    elif form[0] == 'p':
        prob = poly_constrained_primal(f, gts, eqs, p, q, ell, X)
    else:
        raise RuntimeError('Unrecognized form: ' + form + '.')
    return prob


def poly_constrained_primal(f, gts, eqs, p=0, q=1, ell=0, X=None):
    """
    Construct the primal SAGE-(p, q, ell) relaxation for the polynomial optimization problem

        inf{ f(x) : g(x) >= 0 for g in gts,
                    g(x) == 0 for g in eqs,
                    and x in X }

    where :math:`X = R^{\\texttt{f.n}}` by default.

    :param f: a Polynomial.
    :param gts: a list of Polynomials.
    :param eqs: a list of Polynomials.
    :param p: a nonnegative integer.
        Controls the complexity of Lagrange multipliers. p=0 corresponds to scalars.
    :param q: a positive integer.
        The number of folds applied to the constraints "gts" and "eqs". p=1 means "leave gts and eqs as-is."
    :param ell: a nonnegative integer.
        Controls the complexity of any modulator applied to the Lagrangian. ell=0 means that
        the Lagrangian must be SAGE. ell=1 means "tilde_L := modulator * Lagrangian" must be SAGE.
    :param X: None, or a dictionary with three keys 'log_AbK', 'gts', 'eqs' such as that generated by
     the function "conditional_sage_data(...)".

    :return: The primal form SAGE-(p, q, ell) relaxation for the given polynomial optimization problem.
    """
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    lagrangian, ineq_lag_mults, _, gamma = make_poly_lagrangian(f, gts, eqs, p=p, q=q)
    metadata = {'lagrangian': lagrangian}
    if ell > 0:
        alpha_E_q = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
        modulator = Polynomial(2 * alpha_E_q, np.ones(alpha_E_q.shape[0])) ** ell
        lagrangian = lagrangian * modulator
        metadata['modulator'] = modulator
    # The Lagrangian (after possible multiplication, as above) must be a SAGE polynomial.
    con_name = 'Lagrangian sage poly'
    constrs = primal_sage_poly_cone(lagrangian, con_name, log_AbK=X['log_AbK'])
    #  Lagrange multipliers (for inequality constraints) must be SAGE polynomials.
    for s_h, _ in ineq_lag_mults:
        con_name = str(s_h) + ' domain'
        cons = primal_sage_poly_cone(s_h, con_name, log_AbK=X['log_AbK'])
        constrs += cons
    # Construct the coniclifts problem.
    prob = cl.Problem(cl.MAX, gamma, constrs)
    prob.associated_data = metadata
    cl.clear_variable_indices()
    return prob


def poly_constrained_dual(f, gts, eqs, p=0, q=1, ell=0, X=None):
    """
    Construct the dual SAGE-(p, q, ell) relaxation for the polynomial optimization problem

        inf{ f(x) : g(x) >= 0 for g in gts,
                    g(x) == 0 for g in eqs,
                    and x in X }

    where :math:`X = R^{\\texttt{f.n}}` by default.

    :param f: a Polynomial.
    :param gts: a list of Polynomials.
    :param eqs: a list of Polynomials.
    :param p: a nonnegative integer.
        Controls the complexity of Lagrange multipliers. p=0 corresponds to scalars.
    :param q: a positive integer.
        The number of folds applied to the constraints "gts" and "eqs". p=1 means "leave gts and eqs as-is."
    :param ell: a nonnegative integer.
        Controls the complexity of any modulator applied to the Lagrangian. ell=0 means that
        the Lagrangian must be SAGE. ell=1 means "tilde_L := modulator * Lagrangian" must be SAGE.
    :param X: None, or a dictionary with three keys 'log_AbK', 'gts', 'eqs' such as that generated by
     the function "conditional_sage_data(...)".

    :return: The dual SAGE-(p, q, ell) relaxation for the given polynomial optimization problem.
    """
    if X is None:
        X = {'log_AbK': None, 'gts': [], 'eqs': []}
    lagrangian, ineq_lag_mults, eq_lag_mults, _ = make_poly_lagrangian(f, gts, eqs, p=p, q=q)
    metadata = {'lagrangian': lagrangian, 'f': f, 'gts': gts, 'eqs': eqs, 'X': X}
    if ell > 0:
        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
        modulator = Polynomial(2 * alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
        lagrangian = lagrangian * modulator
        f = f * modulator
    else:
        modulator = Polynomial({(0,) * f.n: 1})
    metadata['modulator'] = modulator
    # In primal form, the Lagrangian is constrained to be a SAGE polynomial.
    # Introduce a dual variable "v" for this constraint.
    v = cl.Variable(shape=(lagrangian.m, 1), name='v')
    metadata['v_poly'] = v
    constraints = relative_dual_sage_poly_cone(lagrangian, v, 'Lagrangian', log_AbK=X['log_AbK'])
    for s_g, g in ineq_lag_mults:
        # These generalized Lagrange multipliers "s_g" are SAGE polynomials.
        # For each such multiplier, introduce an appropriate dual variable "v_g", along
        # with constraints over that dual variable.
        v_g = cl.Variable(name='v_' + str(g), shape=(s_g.m, 1))
        constraints += relative_dual_sage_poly_cone(s_g, v_g, name_base=(v_g.name + ' domain'), log_AbK=X['log_AbK'])
        g = g * modulator
        c_g = sym_corr.moment_reduction_array(s_g, g, lagrangian)
        con = c_g @ v == v_g
        con.name += str(g) + ' >= 0'
        constraints.append(con)
    for z_g, g in eq_lag_mults:
        # These generalized Lagrange multipliers "z_g" are arbitrary polynomials.
        # They dualize to homogeneous equality constraints.
        g = g * modulator
        c_g = sym_corr.moment_reduction_array(z_g, g, lagrangian)
        con = c_g @ v == 0
        con.name += str(g) + ' == 0'
        constraints.append(con)
    # Equality constraint (for the Lagrangian to be bounded).
    a = sym_corr.relative_coeff_vector(modulator, lagrangian.alpha)
    constraints.append(a.T @ v == 1)
    # Define the dual objective function.
    obj_vec = sym_corr.relative_coeff_vector(f, lagrangian.alpha)
    obj = obj_vec.T @ v
    # Return the coniclifts Problem.
    prob = cl.Problem(cl.MIN, obj, constraints)
    prob.associated_data = metadata
    cl.clear_variable_indices()
    return prob


def make_poly_lagrangian(f, gts, eqs, p, q):
    """
    Given a problem

    .. math::

        \\begin{align*}
          \min\{ f(x) :~& g(x) \geq 0 \\text{ for } g \\in \\text{gts}, \\\\
                       & g(x) = 0  \\text{ for } g \\in \\text{eqs}, \\\\
                       & \\text{and } x \\in X \}
        \\end{align*}

    construct the q-fold constraints ``q-gts`` and ``q-eqs,`` and the Lagrangian

    .. math::

        L = f - \\gamma
            - \sum_{g \, \\in  \, \\text{q-gts}} s_g \cdot g
            - \sum_{g \, \\in  \, \\text{q-eqs}} z_g \cdot g

    where :math:`\\gamma` and the coefficients on Polynomials :math:`s_g` and :math:`z_g`
    are coniclifts Variables.


    Parameters
    ----------
    f : Polynomial
        The objective in a desired minimization problem.
    gts : list of Polynomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Polynomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    p : int
        Controls the complexity of ``s_g`` and ``z_g``.
    q : int
        The number of folds of constraints ``gts`` and ``eqs``.

    Returns
    -------

    L : Polynomial
        ``L.c`` is an affine expression of coniclifts Variables.

    ineq_dual_polys : a list of pairs of Polynomials.
        If the pair ``(s1, s2)`` is in this list, then ``s1`` is a generalized Lagrange multiplier
        to the constraint ``s2(x) >= 0``.

    eq_dual_polys : a list of pairs of Polynomials.
        If the pair ``(s1, s2)`` is in this list, then ``s1`` is a generalized Lagrange multiplier to the
        constraint ``s2(x) == 0``. This return value is not accessed for primal-form SAGE relaxations.

    gamma : coniclifts.Variable.
        In primal-form SAGE relaxations, we want to maximize ``gamma``. In dual form SAGE relaxations,
        ``gamma`` induces a normalizing equality constraint. This return value is not accessed for
        dual-form SAGE relaxations.

    Notes
    -----
    The values returned by this function are used to construct constrained SAGE relaxations.
    The basic primal SAGE relaxation is obtained by maximizing gamma, subject to the constraint
    that L and each s_g are SAGE polynomials. The dual SAGE relaxation is obtained by symbolically
    applying conic duality to the primal.
    """
    folded_gt = con_gen.up_to_q_fold_cons(gts, q)
    gamma = cl.Variable(name='gamma')
    L = f - gamma
    alpha_E_p = hierarchy_e_k([f] + list(gts) + list(eqs), k=p)
    alpha_multiplier = np.vstack([2 * alpha_E_p, alpha_E_p])
    alpha_multiplier = np.unique(alpha_multiplier, axis=0)
    ineq_dual_polys = []
    for g in folded_gt:
        s_g_coeff = cl.Variable(name='s_' + str(g), shape=(alpha_multiplier.shape[0],))
        s_g = Polynomial(alpha_multiplier, s_g_coeff)
        L -= s_g * g
        ineq_dual_polys.append((s_g, g))
    eq_dual_polys = []
    folded_eq = con_gen.up_to_q_fold_cons(eqs, q)
    for g in folded_eq:
        z_g_coeff = cl.Variable(name='z_' + str(g), shape=(alpha_multiplier.shape[0],))
        z_g = Polynomial(alpha_multiplier, z_g_coeff)
        L -= z_g * g
        eq_dual_polys.append((z_g, g))
    return L, ineq_dual_polys, eq_dual_polys, gamma


def conditional_sage_data(f, gts, eqs):
    """

    Parameters
    ----------
    f : Polynomial
        The objective in a desired optimization problem. This parameter is only used to determine
        the dimension of the set defined by constraints in ``gts`` and ``eqs``.
    gts : list of Polynomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Polynomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.

    Returns
    -------
    X : dict

        ``X`` will be keyed by three strings: ``'log_AbK'``, ``'gts'``, and ``'eqs'``.

        ``X['gts']`` is a list of Polynomials so that every ``g in X['gts']`` has an efficient
        convex representation for ``{log(|x|) : g(|x|) >= 0, |x| > 0}``. (Where the vertical
        bars denote elementwise absolute value, and the logarithm is meant elementwise.)
        The intersection of all of these sets is contained within

                ``{log(|x|) : g(|x|) >= 0 for all g in gts, |x| > 0}``.

        ``X['eqs']`` is defined similarly, but for equality constraints.

        If both ``X['gts']`` and ``X['eqs']`` are empty, then ``X['log_AbK']`` is None.
        Otherwise, ``X['log_AbK']`` is a conic representation of the pointwise, elementwise
        log-absolute-values of the feasible sets cut out by ``X['gts']`` and ``X['eqs']``.
        The conic representation is a triple ``X['log_AbK'] = (A, b, K)``, where ``A`` is a SciPy
        sparse matrix, ``b`` is a numpy 1d array, and ``K`` is a list of coniclifts Cone objects.
        The number of columns for ``A`` in ``X['AbK']`` will always be at least ``f.n``.
        If the number of columns is greater than ``f.n``, then the first f.n columns of ``A``
        correspond (in order!) to the log-absolute-values of variables over which ``f`` is  defined.
        Any remaining columns are auxiliary variables needed to represent ``X`` in coniclifts primitives.

    Notes
    -----

    This function essentially defines the requirements for ``X`` which may be passed to
    conditional SAGE polynomial relaxations defined in this python module.

    It is possible for a user to properly define their own dict ``X`` without calling
    this function. The  only benefit to such an approach is that ``X['gts']`` and ``X['eqs']``
    don't need to be Polynomial objects. As long as ``X['gts']`` and ``X['eqs']`` are callable
    python functions and relate to ``X['log_AbK']`` in the manner described above, then
    you should be able to pass that dict to SAGE relaxations defined in this module
    without trouble. Bear in mind that the functions in ``X['gts']`` and ``X['eqs']`` will
    only be passed elementwise-positive arguments.

    """
    # GP-representable inequality constraints (recast as "Signomial >= 0")
    gp_gts = con_gen.valid_gp_representable_poly_inequalities(gts)
    gp_gts_sigreps = [Signomial(g.alpha_c) for g in gp_gts]
    # GP-representable equality constraints (recast as "Signomial == 0")
    gp_eqs = con_gen.valid_gp_representable_poly_eqs(eqs)
    gp_eqs_sigreps = [Signomial(g.alpha_c) for g in gp_eqs]
    # Fall back on conditional SAGE data implementation for signomials
    dummy_f = Signomial({(0,) * f.n: 1})
    logX = sage_sigs.conditional_sage_data(dummy_f, gp_gts_sigreps, gp_eqs_sigreps)
    X = {'log_AbK': logX['AbK'], 'gts': gp_gts, 'eqs': gp_eqs}
    return X


def hierarchy_e_k(polys, k):
    alpha_tups = sum([list(s.alpha_c.keys()) for s in polys], [])
    alpha_tups = set(alpha_tups)
    s = Polynomial(dict([(a, 1.0) for a in alpha_tups]))
    s = s ** k
    return s.alpha


def log_domain_converter(f):
    fhat = lambda x: f(np.exp(x))
    return fhat



from sageopt.coniclifts.base import Variable, ScalarVariable, Expression
from sageopt.coniclifts.operators.affine import *
from sageopt.coniclifts.operators.relent import relent
from sageopt.coniclifts.operators.exp import weighted_sum_exp
from sageopt.coniclifts.operators.norms import vector2norm
from sageopt.coniclifts.compilers import compile_constrained_system, compile_problem
from sageopt.coniclifts.constraints.constraint import Constraint
from sageopt.coniclifts.constraints.set_membership.product_cone import PrimalProductCone
from sageopt.coniclifts.constraints.set_membership.sage_cone import PrimalSageCone, DualSageCone
from sageopt.coniclifts.constraints.set_membership.conditional_sage_cone import PrimalCondSageCone, DualCondSageCone
from sageopt.coniclifts.problems.problem import Problem
from sageopt.coniclifts.problems.solvers.mosek import Mosek
from sageopt.coniclifts.problems.solvers.ecos import ECOS
from sageopt.coniclifts.standards.constants import minimize as MIN
from sageopt.coniclifts.standards.constants import maximize as MAX
from sageopt.coniclifts.standards.constants import solved as SOLVED
from sageopt.coniclifts.standards.constants import inaccurate as INACCURATE


def clear_variable_indices():
    ScalarVariable._SCALAR_VARIABLE_COUNTER = 0
    # noinspection PyProtectedMember
    Variable._VARIABLE_GENERATION += 1




"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from itertools import product
from collections import defaultdict
import numpy as np
import scipy.sparse as sp

__REAL_TYPES__ = (int, float, np.int_, np.float_, np.float128)


def zero_func():
    return 0


def array_index_iterator(shape):
    return product(*[range(d) for d in shape])


def sparse_matrix_data_to_csc(data_tuples, index_map=None):
    """
    :param data_tuples: a list of quadruplets, each of which contains the data necessary to
    construct a scipy sp matrix.
    :param num_cols: the number of columns to require of the returned matrix (this parameter
    is important if some trailing columns of the matrix are zero, but we want them nevertheless).

    :return: a CSC matrix that is equivalently formed by concatenating the various CSC matrices
    specified by the quadruplets in "data_tuples".
    """
    # d in data_tuples is length 4, and has the following format:
    #   d[0] = A_vals, a list
    #   d[1] = A_rows, a 1d numpy array
    #   d[2] = A_cols, a list
    #   d[3] = the number of rows of this matrix block
    A_cols = []
    A_vals = []
    row_index_offset = 0
    for A_v, A_r, A_c, num_rows in data_tuples:
        A_r[:] += row_index_offset
        A_cols += A_c
        A_vals += A_v
        row_index_offset += num_rows
    A_rows = np.hstack([d[1] for d in data_tuples]).astype(int)
    if index_map is None:
        unique_cols = np.sort(np.unique(A_cols))
        index_map = {c: idx for (idx, c) in enumerate(unique_cols)}
    A_cols = np.array([index_map[ac] for ac in A_cols])
    num_rows = np.max(A_rows) + 1
    num_cols = np.max(list(index_map.values())) + 1
    A = sp.csc_matrix((A_vals, (A_rows, A_cols)),
                      shape=(int(num_rows), int(num_cols)), dtype=float)
    return A, index_map


def parse_cones(K):
    """
    :param K: a list of Cones

    :return: a map from cone type to indices for (A,b) in the conic system
    {x : A @ x + b \in K}, and from cone type to a 1darray of cone lengths.
    """
    m = sum(co.len for co in K)
    type_selectors = defaultdict(lambda: (lambda: np.zeros(shape=(m,), dtype=bool))())
    type_to_cone_start_stops = defaultdict(lambda: list())
    running_idx = 0
    for i, co in enumerate(K):
        type_selectors[co.type][running_idx:(running_idx+co.len)] = True
        type_to_cone_start_stops[co.type] += [(running_idx, running_idx + co.len)]
        running_idx += co.len
    return type_selectors, type_to_cone_start_stops


def sparse_matrix_is_diag(A):
    nonzero_row_idxs, nonzero_col_idxs = A.nonzero()
    nonzero_col_idxs = np.unique(nonzero_col_idxs)
    nonzero_row_idxs = np.unique(nonzero_row_idxs)
    return A.nnz == A.shape[0] and len(nonzero_col_idxs) == len(nonzero_row_idxs)



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
import scipy.sparse as sp
from sageopt.coniclifts import utilities as util
from sageopt.coniclifts.standards import constants as CL_CONSTANTS
from sageopt.coniclifts.problems.solvers.solver import Solver
import copy


class ECOS(Solver):

    @staticmethod
    def apply(c, A, b, K, destructive, compilation_options):
        """
        :return: G, h, cones, A_ecos, b_ecos --- where we expect
         a function call: sol = ecos.solve(c, G, h, cones, A_ecos, b_ecos)
        """
        if not destructive:
            A = A.copy()
            K = copy.deepcopy(K)
        for co in K:
            if co.type not in {'e', 'S', '+', '0'}:
                msg1 = 'ECOS only supports cones with labels in the set {"e", "S", "+", "0"}. \n'
                msg2 = 'The provided data includes an invalid cone labeled "' + str(co[0]) + '".'
                raise RuntimeError(msg1 + msg2)
        type_selectors, type_to_start_stops = util.parse_cones(K)
        # find indices of A corresponding to equality constraints
        A_ecos = A[type_selectors['0'], :]
        b_ecos = -b[type_selectors['0']]  # move to RHS
        # Build "G, h".
        G_nonneg = -A[type_selectors['+'], :]
        h_nonneg = b[type_selectors['+']]
        G_soc = -A[type_selectors['S'], :]
        h_soc = b[type_selectors['S']]
        G_exp = -A[type_selectors['e'], :]
        h_exp = b[type_selectors['e']]
        G = sp.vstack([G_nonneg, G_soc, G_exp], format='csc')
        h = np.hstack((h_nonneg, h_soc, h_exp))
        # create cone dims dict for ECOS
        cones = {'l': int(np.sum(type_selectors['+'])),
                 'e': int(np.sum(type_selectors['e']) / 3),
                 'q': [stop - start for start, stop in type_to_start_stops['S']]}
        data = {'G': G, 'h': h, 'cones': cones, 'A': A_ecos, 'b': b_ecos, 'c': c}
        inv_data = dict()
        return data, inv_data

    @staticmethod
    def solve_via_data(data, params):
        import ecos
        if 'max_iters' in params:
            max_iters = params['max_iters']
        else:
            max_iters = 100000
        sol = ecos.solve(data['c'], data['G'], data['h'], data['cones'], data['A'], data['b'],
                         verbose=params['verbose'],
                         max_iters=max_iters)
        return sol

    # noinspection SpellCheckingInspection
    @staticmethod
    def parse_result(solver_output, inv_data, var_mapping):
        """
        :param inv_data: not used for ECOS.
        :param solver_output: a dictionary returned by the ecos package's "csolve" function.
        The most important fields are "solver_output['x']", which contains the primal optimal
        variable (when it exists), and "solver_output['info']", which contains exit flags
        and solution metrics.

        :param var_mapping: a dictionary mapping names of coniclifts Variables to arrays
        of indices. The array var_mapping['my_var'] contains the indices in
        "solver_output['x']" of the coniclifts Variable object named 'my_var'.

        :return: problem_status, user_variable_values, problem_value. The first of
        these is a string (coniclifts.solved, coniclifts.inaccurate, or coniclifts.failed).
        The second of these is a dictionary from coniclifts Variable names to numpy arrays
        containing the values of those Variables at the returned solution. The last of
        these is either a real number, or +/-np.Inf, or np.NaN.

        Notes: solver_output is described in
            https://github.com/embotech/ecos-python#calling-ecos-from-python
        and solver_output['info'] is described in
            https://www.embotech.com/ECOS/Matlab-Interface/Matlab-Native#MATLAB_API_SolveStatus
        """
        solver_status = solver_output['info']['exitFlag']
        variable_values = dict()
        if solver_status in {0, 1, 2, 10, 11, 12}:
            if solver_status == 0:
                # primal optimal
                problem_status = CL_CONSTANTS.solved
                problem_value = solver_output['info']['pcost']
                variable_values = ECOS.load_variable_values(solver_output['x'], var_mapping)
            elif solver_status == 1:
                # primal infeasible
                problem_status = CL_CONSTANTS.solved
                problem_value = np.Inf
            elif solver_status == 2:
                # dual infeasible (primal unbounded)
                problem_status = CL_CONSTANTS.solved
                problem_value = -np.Inf
            elif solver_status == 10:
                # primal near-optimal
                problem_status = CL_CONSTANTS.inaccurate
                problem_value = solver_output['info']['pcost']
                variable_values = ECOS.load_variable_values(solver_output['x'], var_mapping)
            elif solver_status == 11:
                # primal likely infeasible
                problem_status = CL_CONSTANTS.inaccurate
                problem_value = np.Inf
            else:
                # dual likely infeasible (primal likely unbounded)
                problem_status = CL_CONSTANTS.inaccurate
                problem_value = -np.Inf
        else:
            # solver failed, do not record solution
            #    -1	Maximum number of iterations reached	ECOS_MAXIT
            #    -2	Numerical problems (unreliable search direction)	ECOS_NUMERICS
            #    -3	Numerical problems (slacks or multipliers outside cone)	ECOS_OUTCONE
            #    -4	Interrupted by signal or CTRL-C	ECOS_SIGINT
            #    -7	Unknown problem in solver	ECOS_FATAL
            problem_status = CL_CONSTANTS.failed
            variable_values = dict()
            problem_value = np.NaN
        return problem_status, variable_values, problem_value

    @staticmethod
    def is_installed():
        try:
            import ecos
            return True
        except ImportError:
            return False

    @staticmethod
    def load_variable_values(x, var_mapping):
        """
        :param x: the primal solution returned by ECOS.
        :param var_mapping: a dictionary mapping names of coniclifts Variables to arrays
        of indices. The array var_mapping['my_var'] contains the indices in
        "solver_output['x']" of the coniclifts Variable object named 'my_var'.

        :return: a dictionary mapping names of coniclifts Variables to arrays of their
        corresponding values.
        """
        var_values = dict()
        for var_name in var_mapping:
            var_values[var_name] = x[var_mapping[var_name]]
        return var_values



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt.coniclifts.standards import constants as CL_CONSTANTS
from sageopt.coniclifts.problems.solvers.ecos import ECOS
from sageopt.coniclifts.problems.solvers.mosek import Mosek
from sageopt.coniclifts.compilers import compile_problem
from sageopt.coniclifts.base import Expression
import numpy as np
import time


class Problem(object):

    _SOLVERS_ = {'ECOS': ECOS(), 'MOSEK': Mosek()}

    _SOLVER_ORDER_ = ['MOSEK', 'ECOS']

    def __init__(self, objective_sense, objective_expression, constraints):
        """
        :param objective_sense: either coniclifts.minimize or coniclifts.maximize.
        :param objective_expression: A real numeric type, or Expression with .size == 1, or ScalarExpression.
        :param constraints: a list of coniclifts Constraint objects.
        """
        self.objective_sense = objective_sense
        if not isinstance(objective_expression, Expression):
            objective_expression = Expression(objective_expression)
        self.user_obj = objective_expression
        self.user_cons = constraints
        self.timings = dict()
        t = time.time()
        c, A, b, K, var_name_to_locs, all_vars = compile_problem(objective_expression, constraints)
        if objective_sense == CL_CONSTANTS.minimize:
            self.c = c
        else:
            self.c = -c
        self.timings['compile_time'] = time.time() - t
        self.A = A
        self.b = b
        self.K = K
        self.all_variables = all_vars
        self.user_variable_map = var_name_to_locs
        self.user_variable_values = None
        self.solver_apply_data = dict()
        self.solver_runtime_data = dict()
        self.status = None  # "solved", "inaccurate", or "failed"
        self.value = np.NaN
        self.associated_data = dict()
        self.default_options = {'cache_applytime': False, 'cache_solvetime': False,
                                'destructive': False, 'compilation_options': dict(),
                                'verbose': True}
        pass

    def solve(self, solver=None, **kwargs):
        """
        :param solver: a string. 'MOSEK' or 'ECOS'.
        :return:
        """
        if solver is None:
            for svr in Problem._SOLVER_ORDER_:
                if Problem._SOLVERS_[svr].is_installed():
                    solver = svr
                    break
        if solver is None:
            raise RuntimeError('No acceptable solver is installed.')
        options = self.default_options.copy()
        options.update(kwargs)
        solver_object = Problem._SOLVERS_[solver]
        if not solver_object.is_installed():
            raise RuntimeError('Solver "' + solver + '" is not installed.')
        self.timings[solver] = dict()

        # Finish solver-specific compilation
        t0 = time.time()
        data, inv_data = solver_object.apply(self.c, self.A, self.b, self.K,
                                             options['destructive'],
                                             options['compilation_options'])
        self.timings[solver]['apply'] = time.time() - t0
        if options['cache_applytime']:
            self.solver_apply_data[solver] = data

        # Solve the problem
        t1 = time.time()
        raw_result = solver_object.solve_via_data(data, options)
        self.timings[solver]['solve_via_data'] = time.time() - t1
        if options['cache_solvetime']:
            self.solver_runtime_data[solver] = raw_result
        parsed_result = solver_object.parse_result(raw_result, inv_data, self.user_variable_map)
        self.status = parsed_result[0]
        self.user_variable_values = parsed_result[1]

        # Load values into ScalarVariable objects.
        if len(self.user_variable_values) > 0:
            for v in self.all_variables:
                if v.name in self.user_variable_values:
                    var_val = self.user_variable_values[v.name]
                    v.set_scalar_variables(var_val)

        # Record objective value
        if self.status in {CL_CONSTANTS.solved, CL_CONSTANTS.inaccurate}:
            if self.objective_sense == CL_CONSTANTS.minimize:
                self.value = parsed_result[2]
            else:
                self.value = -parsed_result[2]
        else:
            self.value = np.NaN

        self.timings[solver]['total'] = time.time() - t0
        return self.status, self.value








"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.base import NonlinearScalarAtom, Expression, ScalarExpression, Variable
from sageopt.coniclifts.cones import Cone


def vector2norm(x):
    if not isinstance(x, Expression):
        x = Expression(x)
    x = x.ravel()
    d = {Vector2Norm(x): 1}
    return ScalarExpression(d, 0).as_expr()


class Vector2Norm(NonlinearScalarAtom):

    _VECTOR_2_NORM_COUNTER_ = 0

    @staticmethod
    def __atom_text__():
        return 'Vector2Norm'

    def __init__(self, args):
        args = args.as_expr().ravel()
        self._args = tuple(self.parse_arg(v) for v in args)
        self._id = Vector2Norm._VECTOR_2_NORM_COUNTER_
        Vector2Norm._VECTOR_2_NORM_COUNTER_ += 1
        v = Variable(shape=(), name='_vec2norm_epi[' + str(self.id) + ']_')
        self._epigraph_variable = v[()].scalar_variables()[0]

    def is_convex(self):
        return True

    def is_concave(self):
        return False

    def epigraph_conic_form(self):
        """
        Generate conic constraint for epigraph
            np.linalg.norm( np.array(self.args), ord=2) <= self._epigraph_variable
        The coniclifts standard for the second order cone (of length n) is
            { (t,x) : x \in R^{n-1}, t \in R, || x ||_2 <= t }.
        """
        m = len(self.args) + 1
        b = np.zeros(m,)
        A_rows, A_cols, A_vals = [0], [self._epigraph_variable.id], [1]  # for first row
        for i, arg in enumerate(self.args):
            A_rows += (len(arg)-1) * [i + 1]
            for var, coeff in arg[:-1]:
                A_cols.append(var.id)
                A_vals.append(coeff)
            b[i+1] = arg[-1][1]
        K = [Cone('S', m)]
        A_rows = np.array(A_rows)
        return A_vals, A_rows, A_cols, b, K

    def value(self):
        raise NotImplementedError()




"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
import scipy.sparse as sp
from sageopt.coniclifts import utilities as util
from collections import defaultdict
from sageopt.coniclifts.cones import Cone


def build_cone_type_selectors(K):
    """
    :param K: a list of Cones

    :return: a map from cone type to indices for (A,b) in the conic system
    {x : A @ x + b \in K}, and from cone type to a 1darray of cone lengths.
    """
    m = sum(co.len for co in K)
    type_selectors = defaultdict(lambda: (lambda: np.zeros(m, dtype=bool))())
    running_idx = 0
    for i, co in enumerate(K):
        type_selectors[co.type][running_idx:(running_idx+co.len)] = True
        running_idx += co.len
    return type_selectors


def find_cones_with_mutually_disjoint_scope(cones):
    # The runtime of this function scales quadratically in the length of the list "cones".
    # If Mosek is going to be viable for compiling cone programs of nontrivial size, then
    # this function needs to be orders of magnitude faster. In addition to changing how it's
    # written, the final function might be simple enough to be accelerated by numba or cython.
    disjoint_cone_idxs = set()
    remaining_cone_idxs = {i for i in range(len(cones))}
    for co_idx, co in enumerate(cones):
        if 'isolated' in co.annotations and co.annotations['isolated']:
            disjoint_cone_idxs.add(co_idx)
            remaining_cone_idxs.remove(co_idx)
        else:
            co_cols = co.annotations['scope']
            is_isolated = True
            for other_idx in remaining_cone_idxs:
                if other_idx == co_idx:
                    continue
                other_cols = cones[other_idx].annotations['scope']
                if len(co_cols.intersection(other_cols)) > 0:
                    is_isolated = False
                    break
            if is_isolated:
                disjoint_cone_idxs.add(co_idx)
                remaining_cone_idxs.remove(co_idx)
    disjoint_cones = []
    for i in disjoint_cone_idxs:
        disjoint_cones.append(cones[i])
    return disjoint_cones


def find_diagonal_cone_constraints(A, K):
    A = sp.csr_matrix(A)
    diag_cones = []
    for co in K:
        co_start, co_stop = co.annotations['ss']
        # Check if this cone is diagonal.
        curr_A = A[co_start:co_stop, :].tocoo()
        if 'diagonal' in co.annotations and co.annotations['diagonal']:
            co.annotations['A'] = curr_A
            diag_cones.append(co)
        elif util.sparse_matrix_is_diag(curr_A):
            co.annotations['A'] = curr_A
            diag_cones.append(co)
    return diag_cones


def prep_sep_dis_diag_cone_cons(A, b, K, dont_sep=None):
    """
    "Prepare to separate disjoint diagonal cone constraints (from the conic system (A, b, K))."

    K is a list a Cone objects.

    """
    if dont_sep is None:
        dont_sep = set()
    dont_sep = {'0'}.union(dont_sep)
    #
    # Before doing anything we annotate the cones with the information necessary
    # to place themselves relative to (A,b). Then we find the disjoint diagonal cones.
    #
    Cone.annotate_cone_positions(K)  # gives cones 'ss' and 'position' annotations
    relevant_cones = [co for co in K if co.type not in dont_sep]
    diag_cones = find_diagonal_cone_constraints(A, relevant_cones)  # gives cones an 'A' annotation
    Cone.annotate_cone_scopes(A, diag_cones)  # gives cones a 'scope' annotation.
    disjoint_diag_cones = find_cones_with_mutually_disjoint_scope(diag_cones)
    disjoint_diagonal_idxs = {co.annotations['position'] for co in disjoint_diag_cones}
    #
    # Build the return values
    #
    K0 = [co for i, co in enumerate(K) if i not in disjoint_diagonal_idxs]
    for i in range(len(K0)):
        K0[i].annotations = dict()
    sep_K0 = []
    type_selectors = build_cone_type_selectors(K)
    scalings = np.ones(A.shape[1])
    translates = np.zeros(A.shape[1])
    for co in disjoint_diag_cones:
        co_start, co_stop = co.annotations['ss']
        # record normalization and translation necessary to homogenize these cones
        curr_A = co.annotations['A']  # curr_A is COO format.
        nz_vals, nz_rows, nz_cols = curr_A.data, curr_A.row, curr_A.col
        for i, col_idx in enumerate(nz_cols):
            row_idx = co_start + nz_rows[i]
            translates[col_idx] -= b[row_idx]
            scalings[col_idx] /= nz_vals[i]
        # record a map from coordinate indices in this this cone to the columns
        # (variable indices) that must belong to the cone at the coordinate index.
        # i.e., sort the entries of nz_cols according to the entries of nz_rows
        type_selectors[co.type][co_start:co_stop] = False
        col_mapping = nz_cols[np.argsort(nz_rows)]
        new_co = Cone(co.type, co.len, {'col mapping': col_mapping})
        sep_K0.append(new_co)
    # Build the final selector
    all_selectors = [sel.reshape(-1, 1) for sel in type_selectors.values()]
    all_selectors = np.hstack(all_selectors)
    row_selector = np.any(all_selectors, axis=1)
    return row_selector, K0, sep_K0, scalings, translates


def separate_disjoint_diagonal_cones(A, b, K, destructive=False, dont_sep=None):
    """
    This function's purpose ...

        coniclifts' standard form for a feasible set is {x : A @ x + b \in K}. Some solvers
        require alternative standard forms, such as {x : A @ x == b, x \in K },
        or {x : A @ x <= b, x \in K }. Mosek requires an even more peculiar form, namely
        {(x,X) : A @ x + F(X) <= b, x \in K, X >> 0 }, where F is a particular linear map.

        We can trivially convert from the coniclifts standard form to the above forms by
        introduction of slack variables. I.e. represent {x : A @ x + b \in K} as
        { x : y == A @ x + b, y \in K }. The drawback is that doing this substitution
        can substantially worsen the conditioning of the problem. This function exists
        to facilitate similar reformulation, without introducing unnecessary slack
        variables.

    :param A: a scipy csc sparse matrix (m rows, n columns)
    :param b: a numpy 1darray of length m
    :param K: a list of cones, specified by (str, int) pairs, where "str" gives the
    cones type, and "int" gives the cone's length.
    :param destructive: a boolean. Indicates whether or not inputs (A,b) are modified,
    or copied.
    :param dont_sep: a set of cone types that are not to be modified while reformulating
    the system. Common values are dont_sep={'0'} and dont_sep={'0','+'}.
    :return: See in-line comments.
    """
    if not destructive:
        A = A.copy()
        b = b.copy()
    sel, K0, sep_K0, scale, trans = prep_sep_dis_diag_cone_cons(A, b, K, dont_sep=dont_sep)
    # In order to separate disjoint diagonal cones from the system (A, b, K) as suggested by
    # the function above, we will need to apply a diagonal affine change of coordinates to
    # {x : A @ x + b \in K}. That diagonal change of coordinates maps "x" to
    # x0 == np.diag(scale) @ (x + trans).
    #
    # Since "x" and "x0" don't yet exist on the computer, we transform (A, b) to
    # (A0, b0) in such a way that the set of feasible "x" maps to {x0 : A0 @ x0 + b0 \in K}.
    b0 = b + A.dot(trans)
    A0 = A @ sp.diags(scale)
    # Now that this map has been applied, we can drop a subset of rows from (A, b)
    # --- they are now accounted for by "sep_K0". The returned value "K0" has
    # already been updated to reflect the fact that we intend to drop these rows.
    A0 = A0[sel, :]
    b0 = b0[sel]
    # Our feasible set has now been mapped to { x0 : A0 @ x0 + b0 \in K0 and x0 \in sep_K0 }.
    #
    # In order to invert this map later, we return both (A0, b0, K0, sep_K0) and the
    # transformation data given by (scale, trans).
    return A0, b0, K0, sep_K0, scale, trans


def replace_nonzero_cones_with_zero_cones_and_slacks(A, K, destructive=False, dont_rep=None):
    # It only really makes sense to call this function after any isolated diagonal cone
    # constraints have been separated / factored out of (A,K).
    if not destructive:
        A = A.copy()
        K = K.copy()
    if dont_rep is None:
        dont_rep = set()
    allowed = {'0'}.union(dont_rep)  # the zero cone is never replaced by slack variables
    running_row_idx = 0
    running_new_var_idx = 0
    slacks_K = []
    aug_data = [[], [], []]  # the constituent lists will store values, rows, and columns
    for i in range(len(K)):
        (co_type, co_len) = K[i].type, K[i].len
        if co_type not in allowed:
            new_col_idxs = np.arange(running_new_var_idx, running_new_var_idx + co_len)
            running_new_var_idx += co_len
            aug_data[2].append(new_col_idxs)
            new_slack_co = Cone(co_type, co_len, {'col mapping': A.shape[1] + new_col_idxs})
            slacks_K.append(new_slack_co)
            aug_data[1].append(np.arange(running_row_idx, running_row_idx + co_len))
            K[i] = Cone('0', co_len)
        running_row_idx += co_len
    if len(aug_data[2]) > 0:
        aug_data[2] = np.hstack(aug_data[2])
        aug_data[1] = np.hstack(aug_data[1])
        aug_data[0] = -1 * np.ones(len(aug_data[1]))
        augmenting_matrix = sp.csc_matrix((aug_data[0], (aug_data[1], aug_data[2])),
                                          shape=(A.shape[0], running_new_var_idx))
        A = sp.hstack([A, augmenting_matrix], format='csc')
    return A, K, slacks_K


def separate_cone_constraints(A, b, K, destructive=False, dont_sep=None, avoid_slacks=False):
    """
    Replace the conic system {x : A @ x + b \in K } by {x0 : A1 @ x1 + b1 \in K}

    :param A: a scipy csc sparse matrix (m rows, n columns)
    :param b: a numpy 1darray of length m
    :param K: a list of cones, specified by (str, int) pairs, where "str" gives the
    cones type, and "int" gives the cone's length.
    :param destructive: a boolean. Indicates whether or not inputs (A,b) are modified,
    or copied.
    :param dont_sep: a set of cone types that are not to be modified while reformulating
    the system. Common values are dont_sep={'0'} and dont_sep={'0','+'}.
    :param avoid_slacks: boolean. Indicates whether or not to perform pre-processing to minimize
    the number of slack variables.

    :return: See in-line comments.

    Example: Suppose the conic system (A,b,K) can be expressed as {x : B @ x == d, G @ x + h >=0, x >=0 }, where G
             is a non-square matrix. Then by calling this function with dont_sep=None, we obtain the conic system
                { [x,y] : A1 @ [x, y] == b0 } \cap {[x, y] : x >= 0, y >= 0}
             where A1 = [[B, 0], [G, -I]] and b0 = [d, -h]. This conic system is equivalent to (A,b,K) once we
             project onto the x coordinates.

             This case is interesting because it shows that a cone constraint "x >= 0" can be separated from
             from the matrix system, even though "x" appears in th inequality constraint "G @ x + d >= 0".

    Example: Suppose the conic system (A,b,K) can be expressed as {x : B @ x == d, L <= x <= U }.
             Then by calling this function with dont_sep=None, we obtain the conic system
                { [x,y,z] : A1 @ [x,y,z] == b0 } \cap {[x,y,z] : y >= 0, z >= 0 }
             where A1 = [[B, 0, 0], [I, -I, 0], [-I, 0, -I]] and b0 = [b, L, -U]. This conic system is equivalent
             to (A,b,K) once we project out the [y,z] components.

             NOTE: here we ended up introducing slack variables for both inequalities in the variable "x", rather
             than writing {[x,w] : B @ x == b + B @ L, U - L - x - w == 0} \cap {[x,w] : x >= 0, w >= 0 }. This
             Second option is still a valid formulation, but it is not one that this function would generate.

    Example: Suppose the conic system (A,b,K) can be expressed as  { x : B @ x == d, G @ x <=_{K_prime} h },
             where every entry of G is nonzero, A is not square, and K_prime does not contain the zero cone.
             Then dont_sep={'0'} yields
                { [x, y] : A1 @ [x, y] == b0 } \cap { [x, y] : y \in K_{prime} }
             for A1 = [[B, 0], [-G, -I]] and b0 = [d, -h].

    Common use-case: Passing dont_sep={'0','+'}. This ensures that any constraints on nonlinear cones
            (i.e. exp, psd, soc, etc...) are stated simply as "x[indices_j] \in K_j", where
            "indices_{j1}" and "indices_{j2}" are disjoint whenever j1 \neq j2. This is the main
            modeling requirement for MOSEK's standard form.

    Alternative use-case: Passing dont_sep={'+','0','e','S'} for a conic system involving semidefinite
            constraints. This will force all of the "SDP-ness" of the conic system into disjoint slack variables,
            while aspects of the conic system regarding other cones (e.g. second order cones, exponential cones,
            nonnegative orthants, etc...) are left alone.

    REMARKS:
        (1) dont_sep=None is equivalent to dont_sep={'0'}.
    """
    if not destructive:
        A = A.copy()
        K = K.copy()
    if dont_sep is None:
        dont_sep = set('0')
    if avoid_slacks:
        # (1) identify "simple" nonlinear cones, and factor them out from the matrix part of the conic system.
        A0, b0, K0, sep_K0, scale, trans = separate_disjoint_diagonal_cones(A, b, K, True, dont_sep)
        #   A0 has the same number of columns as A, but likely fewer rows.
        #   K0 is no longer than K.
        #   sep_K0 contains information for block elementwise constraints on "x".
        #   scale and trans contain the data for a diagonal affine change of coordinates between our
        #   original variables {x : A @ x + b \in K} and the new variables {x0 : A0 @ x0 + b0 \in K0, x0 \in sep_K0 }.
    else:
        A0 = A
        K0 = K
        b0 = b
        sep_K0 = []
        scale = np.ones(A.shape[1])
        trans = np.zeros(A.shape[1])
    # (2) introduce slack variables (and separately record constraints) to factor remaining constraints
    #     of type "dont_sep" out of the matrix and into disjoint diagonal homogeneous cone constraints.
    A1, K1, sep_K1 = replace_nonzero_cones_with_zero_cones_and_slacks(A0, K0, destructive=True, dont_rep=dont_sep)
    scale = np.hstack((scale, np.ones(A1.shape[1] - len(scale))))
    trans = np.hstack((trans, np.zeros(A1.shape[1] - len(trans))))
    #   A1 has the same number of rows of A0, but likely has more columns.
    #   K1 is of the same length of K0, but only contains the cones in "dont_sep".
    #   sep_K1 defines the constraints on any slack variables which allowed us to replace K0 by K1 as above.
    #   The introduction of new variables required that we appropriately pad "scale" and "trans".
    sep_K = sep_K0 + sep_K1
    return A1, b0, K1, sep_K, scale, trans


def separated_cones_to_matrix_cones(sep_K, num_cols, destructive=False):
    m = sum(co.len for co in sep_K)
    A_rows = np.arange(m)
    A_vals = np.ones(m)
    A_cols = np.hstack(co.annotations['col mapping'] for co in sep_K)
    if not destructive:
        # overwrite the reference to "sep_K" because we no longer need the annotations.
        sep_K = [Cone(co.type, co.len) for co in sep_K]
    else:
        for co in sep_K:
            co.annotations = dict()
    b = np.zeros(m)
    A = sp.csc_matrix((A_vals, (A_rows, A_cols)), shape=(m, num_cols), dtype=float)
    return A, b, sep_K



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt import coniclifts
from sageopt import relaxations
from sageopt import symbolic

from sageopt.relaxations import local_refine, local_refine_polys_from_sigs
from sageopt.relaxations import conditional_sage_data
from sageopt.relaxations import sage_feasibility, sage_multiplier_search

from sageopt.symbolic.signomials import Signomial, standard_sig_monomials
from sageopt.relaxations import sig_primal, sig_dual
from sageopt.relaxations import sig_constrained_primal, sig_constrained_dual
from sageopt.relaxations import sig_relaxation, sig_constrained_relaxation
from sageopt.relaxations import sig_solrec

from sageopt.symbolic.polynomials import Polynomial, standard_poly_monomials
from sageopt.relaxations import poly_primal, poly_dual
from sageopt.relaxations import poly_constrained_primal, poly_constrained_dual
from sageopt.relaxations import poly_relaxation, poly_constrained_relaxation
from sageopt.relaxations import poly_solrec

__version__ = '0.3.4'



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import numpy as np
import sageopt.coniclifts as cl
from collections import defaultdict
from sageopt.symbolic.signomials import Signomial

__NUMERIC_TYPES__ = (int, float, np.int_, np.float_)


def standard_poly_monomials(n):
    """
    Parameters
    ----------
    n : int
        The dimension of the space over which the constituent Polynomials will be defined.

    Returns
    -------
    x : NumPy ndarray
        An array  of length ``n``. ``x[i]`` is the Polynomial that evaluates to ``x[i](z) = z[i]`` for all numpy
        ndarrays ``z`` of length ``n``.

    Example
    -------
    This function is useful for constructing Polynomials in an algebraic form. ::

        x = standard_poly_monomials(3)
        f = (x[0] ** 2) * x[2] - 4 * x[1] * x[2] * x[0]

    """
    x = np.empty(shape=(n,), dtype=object)
    for i in range(n):
        ei = np.zeros(shape=(1, n))
        ei[0, i] = 1
        x[i] = Polynomial(ei, np.array([1]))
    return x


class Polynomial(Signomial):
    """
    This class provides a symbolic representation for polynomials which are sparse in the monomial basis.
    These functions look like the following.

    .. math::

       x \mapsto \sum_{i=1}^m c_i \prod_{j=1}^n {x_j}^{\\alpha_{ij}}

    The constructor for this class can be called in two different ways. The arguments
    to the constructor have names which reflect the different possibilities.

    Parameters
    ----------

    alpha_maybe_c : dict or NumPy ndarray

         If ``alpha_maybe_c`` is a dict, then it must be a dictionary from tuples-of-numbers to
         scalars. The keys will be converted to rows of a matrix which we call ``alpha``, and
         the values will be assembled into a vector which we call ``c``.

         If ``alpha_maybe_c`` is a NumPy ndarray, then the argument ``c`` must also be an ndarray,
         and ``c.size`` must equal ``alpha_maybe_c.shape[0]``.

    c : None or NumPy ndarray

        This value is only used when ``alpha_maybe_c`` is a NumPy ndarray. If that is the case, then
        this Polynomial will represent the function ``lambda x: c @ np.prod(np.power(alpha_maybe_c, x))``.

    Examples
    --------

    There are two ways to call the Polynomial constructor.

    The first way is to specify a dictionary from tuples to scalars. Each key-value pair of the dictionary
    represents a monomial appearing in this polynomial. The key tuples must all be of some common length ``n``.
    If ``(a, coeff)`` is a key-value pair in this dictionary, then the Polynomial includes an additive term
    ``lambda x: coeff * np.prod(np.power(a, x))``. ::

        alpha_and_c = {(1,): 2}
        f = Polynomial(alpha_and_c)
        print(f(1))  # equal to 2.
        print(f(-3))  # equal to -6.

    The second way is to specify two arguments. In this case the first argument ``alpha`` is a NumPy array of
    exponent vectors, where ``alpha[i, j]`` is the power of variable ``j`` in monomial ``i``. The second argument
    is a numpy array ``c``, where ``c[i]`` is the coefficient on the i-th monomial defined by ``alpha``. ::

        alpha = np.array([[1, 0], [0, 1], [1, 1]])
        c = np.array([1, 2, 3])
        f = Polynomial(alpha, c)
        x = np.array([-4, 7])
        val = f(x)  # val = 1 * (-4) + 2 * (7) + 3 * (-4 * 7)
        print(val)  # -74

    Attributes
    ----------

    n : int
        The dimension of the space over which this Polynomial is defined. The number of columns in ``alpha``,
        and the length of tuples appearing in the dictionary ``alpha_c``.

    m : int
        The number of terms needed to express this Polynomial in the standard monomial basis.
        The number of rows in  ``alpha``.  The length of the dictionary ``alpha_c``.

    alpha : NumPy ndarray
        Has shape ``(m, n)``. Entry ``alpha[i,j]`` is the power of an implicit variable ``x[j]``
        appearing in the i-th monomial for this Polynomial. The i-th monomial, in turn, has coefficient ``c[i]``.

    c : NumPy ndarray
        Has shape ``(m,)``. The scalar ``c[i]`` is this Polynomial's coefficient on the basis function
        ``lambda x: np.prod(np.power(alpha[i, :], x))``. It is possible to have ``c.dtype == object``, to allow for
        coniclifts Variables.

    alpha_c : defaultdict
        The keys of ``alpha_c`` are tuples of length ``n``, containing real numeric types (e.g int, float).
        These tuples correspond to rows in ``alpha``.

    Notes
    -----
    The Polynomial class subclasses the Signomial class. This is done because most algebraic operations between
    polynomials and signomials are identical. However it is important to remember that polynomials and signomials
    evaluate in very different ways.

    """

    def __init__(self, alpha_maybe_c, c=None):
        Signomial.__init__(self, alpha_maybe_c, c)
        if not np.all(self.alpha % 1 == 0):
            raise RuntimeError('Exponents must belong the the integer lattice.')
        if not np.all(self.alpha >= 0):
            raise RuntimeError('Exponents must be nonnegative.')
        self._sig_rep = None
        self._sig_rep_constrs = []
        pass

    @property
    def grad(self):
        """
        A numpy ndarray of shape ``(n,)`` whose entries are Polynomials. For a numpy ndarray ``x``,
        ``grad[i](x)`` is the partial derivative of this Polynomial with respect to coordinate ``i``,
        evaluated at ``x``. This array is constructed only when necessary, and is cached upon construction.
        """
        Signomial._cache_grad(self)
        return self._grad

    @property
    def hess(self):
        """
        A numpy ndarray of shape ``(n, n)``, whose entries are Polynomials. For a numpy ndarray ``x``,
        ``hess[i,j](x)`` is the (i,j)-th partial derivative of this Polynomial, evaluated at ``x``.
        This array is constructed only when necessary, and is cached upon construction.
        """
        Signomial._cache_hess(self)
        return self._hess

    def __mul__(self, other):
        if not isinstance(other, Polynomial):
            if isinstance(other, Signomial):
                raise RuntimeError('Cannot multiply signomials and polynomials.')
            # else, we assume that "other" is a scalar type
            tup = (0,) * self.n
            d = {tup: other}
            other = Polynomial(d)
        self_var_coeffs = (self.c.dtype not in __NUMERIC_TYPES__)
        other_var_coeffs = (other.c.dtype not in __NUMERIC_TYPES__)
        if self_var_coeffs and other_var_coeffs:
            raise RuntimeError('Cannot multiply two polynomials that contain non-numeric coefficients.')
        temp = Signomial.__mul__(self, other)
        temp = temp.as_polynomial()
        return temp

    def __truediv__(self, other):
        if not isinstance(other, __NUMERIC_TYPES__):
            raise RuntimeError('Cannot divide a polynomial by the non-numeric type: ' + type(other) + '.')
        other_inv = 1 / other
        return self.__mul__(other_inv)

    def __add__(self, other):
        if isinstance(other, Signomial) and not isinstance(other, Polynomial):
            raise RuntimeError('Cannot add signomials to polynomials.')
        temp = Signomial.__add__(self, other)
        temp = temp.as_polynomial()
        return temp

    def __sub__(self, other):
        if isinstance(other, Signomial) and not isinstance(other, Polynomial):
            raise RuntimeError('Cannot subtract a signomial from a polynomial (or vice versa).')
        temp = Signomial.__sub__(self, other)
        temp = temp.as_polynomial()
        return temp

    def __rmul__(self, other):
        # multiplication is commutative
        return Polynomial.__mul__(self, other)

    def __radd__(self, other):
        # addition is commutative
        return Polynomial.__add__(self, other)

    def __rsub__(self, other):
        # subtract self, from other
        # rely on correctness of __add__ and __mul__
        # noinspection PyTypeChecker
        return other + (-1) * self

    def __neg__(self):
        # rely on correctness of __mul__
        # noinspection PyTypeChecker
        return (-1) * self

    def __pow__(self, power, modulo=None):
        if self.c.dtype not in __NUMERIC_TYPES__:
            raise RuntimeError('Cannot exponentiate polynomials with symbolic coefficients.')
        temp = Signomial(self.alpha_c)
        temp = temp ** power
        temp = temp.as_polynomial()
        return temp

    def __call__(self, x):
        """
        Parameters
        ----------
        x : NumPy ndarray
            This vector can contain real numeric types and Polynomial objects.

        Returns
        -------
        val : float or Polynomial
            If ``x`` is purely numeric, then ``val`` is a float.
            If ``x`` contains Polynomial objects, then ``val`` is a Polynomial object.

        Examples
        --------
        Evaluating a Polynomial at a numeric point. ::

            p = Polynomial({(1,2,3): -1})
            x = np.array([3,2,1])
            print(p(x))  #  (3 ** 1) * (2 ** 2) * (1 ** 3) * (-1) == -12

        Evaluating a Polynomial on another polynomial. ::

            p = Polynomial({(2,): 1})  # represents lambda x: x ** 2
            z = Polynomial({(1,): 2, (0,): -1})  # represents lambda x: x - 1
            w = p(z)  # represents lambda x: (x - 1) ** 2
            print(w(1))  # equals zero
            print(w(2))  # equals one
            print(w(0))  # equals one

        Notes
        -----
        If ``x`` contains Polynomial entries, then those Polynomials must all be over the same
        number of variables. However, those Polynomials need not have the same number of variables
        as the current polynomial.

        """
        if np.isscalar(x) or (isinstance(x, np.ndarray) and x.size == 1) or isinstance(x, Polynomial):
            # noinspection PyTypeChecker
            x = np.array([np.asscalar(np.array(x))])  # coerce into a 1d array of shape (1,).
        if not x.shape[0] == self.n:
            raise ValueError('The point must be in R^' + str(self.n) +
                             ', but the provided point is in R^' + str(x.shape[0]))
        if x.dtype == 'object':
            ns = np.array([xi.n for xi in x.flat if isinstance(xi, Polynomial)])
            if ns.size == 0:
                x = x.astype(np.float)
            elif np.any(ns != ns[0]):
                raise RuntimeError('The input vector cannot contain Polynomials over different variables.')
        if x.ndim == 1:
            temp1 = np.power(x, self.alpha)
            temp2 = np.prod(temp1, axis=1)
            val = np.dot(self.c, temp2)
            return val
        elif x.ndim == 2:
            vals = [self.__call__(xi) for xi in x.T]
            val = np.array(vals)
            return val
        else:
            raise ValueError('Can only evaluate on x with dimension <= 2.')

    def __hash__(self):
        return hash(frozenset(self.alpha_c.items()))

    def __eq__(self, other):
        if not isinstance(other, Polynomial):
            return False
        if self.m != other.m:
            return False
        for k in self.alpha_c:
            v = self.alpha_c[k]
            other_v = other.query_coeff(np.array(k))
            if not cl.Expression.are_equivalent(other_v, v, rtol=1e-8):
                return False
        return True

    def remove_terms_with_zero_as_coefficient(self):
        """
        Update ``alpha``, ``c``, and ``alpha_c`` to remove nonconstant terms where ``c[i] == 0``.
        """
        Signomial.remove_terms_with_zero_as_coefficient(self)
        pass

    def query_coeff(self, a):
        """
        Returns the coefficient of the monomial ``lambda x: np.prod(np.power(a, x))`` for this Polynomial.
        """
        tup = tuple(a)
        if tup in self.alpha_c:
            return self.alpha_c[tup]
        else:
            return 0

    def constant_location(self):
        """
        Return the index ``i`` so that ``alpha[i, :]`` is the zero vector.
        """
        loc = Signomial.constant_location(self)
        return loc

    def even_locations(self):
        """
        Return the largest ndarray ``evens``, so that ``np.all(alpha[evens,:] % 2 == 0)``.
        """
        evens = np.where(~(self.alpha % 2).any(axis=1))[0]
        return evens

    def partial(self, i):
        """
        Compute the symbolic partial derivative of this polynomial, at coordinate ``i``.

        Parameters
        ----------
        i : int
            ``i`` must be an integer from 0 to ``self.n - 1``.

        Returns
        -------
        p : Polynomial
            The function obtained by differentiating this polynomial with respect to its i-th argument.
        """
        if i < 0 or i >= self.n:
            raise RuntimeError('This polynomial does not have an input at index ' + str(i) + '.')
        d = defaultdict(int)
        for j in range(self.m):
            if self.alpha[j, i] > 0:
                vec = self.alpha[j, :].copy()
                c = self.c[j] * vec[i]
                vec[i] -= 1
                d[tuple(vec.tolist())] = c
        d[self.n * (0,)] += 0
        p = Polynomial(d)
        return p

    def standard_multiplier(self):
        """
        Returns a Polynomial which is globally nonnegative by construction, for use as a modulator in SAGE hierarchies.
        The particular polynomial has exponents ``alpha = alpha[even_locations(), :]``, and a coefficient vector of
        all ones.
        """
        evens = self.even_locations()
        mult_alpha = self.alpha[evens, :].copy()
        mult_c = np.ones(len(evens))
        mult = Polynomial(mult_alpha, mult_c)
        if mult.alpha_c[self.n * (0,)] == 0:
            mult += 1
        return mult

    @property
    def sig_rep(self):
        """
        Return the signomial representative of the current Polynomial, as well as a list of constraints needed
        to enforce the relationship between the current Polynomial and the signomial representative.

        Returns
        -------
        sr : Signomial
            If this Signomial is globally nonnegative, then the current Polynomial is also globally nonnegative.

        sr_cons : list of coniclifts Constraints
            If the current Polynomial has nonconstant coefficients (i.e. some entries of ``c`` are coniclifts
            Variables), then ``sr`` will also have nonconstant coefficients. In order to enforce the relationship
            between ``sr`` and the current Polynomial, we may require constraints between ``c`` and ``sr.c``.
            Any such constraints are in this list.

        """
        # It is important that self._sig_rep.alpha == self.alpha.
        if self._sig_rep is None:
            self._compute_sig_rep()
        sr = self._sig_rep
        sr_cons = self._sig_rep_constrs
        return sr, sr_cons

    def _compute_sig_rep(self):
        self._sig_rep = None
        self._sig_rep_constrs = []
        sigrep_c = np.zeros(shape=(self.m,), dtype=object)
        need_vars = []
        for i, row in enumerate(self.alpha):
            if np.any(row % 2 != 0):
                if isinstance(self.c[i], __NUMERIC_TYPES__):
                    sigrep_c[i] = -abs(self.c[i])
                else:
                    need_vars.append(i)
            else:
                if isinstance(self.c[i], np.ndarray):
                    sigrep_c[i] = self.c[i][()]
                else:
                    sigrep_c[i] = self.c[i]
        if len(need_vars) > 0:
            var_name = str(self) + ' variable sigrep coefficients'
            c_hat = cl.Variable(shape=(len(need_vars),), name=var_name)
            sigrep_c[need_vars] = c_hat
            self._sig_rep_constrs.append(c_hat <= self.c[need_vars])
            self._sig_rep_constrs.append(c_hat <= -self.c[need_vars])
        self._sig_rep = Signomial(self.alpha, sigrep_c)
        pass

    def as_signomial(self):
        """
        Returns
        -------
        f : Signomial
            For every elementwise positive vector ``x``, we have ``self(x) == f(np.log(x))``.
        """
        f = Signomial(self.alpha_c)
        return f



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.base import Expression


def cast_res(res):
    if isinstance(res, np.ndarray) and res.dtype == object:
        return Expression(res)  # includes Variables
    else:
        return res  # should only be Expressions, ScalarExpressions, and numeric ndarrays.

# Basic mathematical functions
#   https://docs.scipy.org/doc/numpy/reference/routines.math.html


# noinspection PyShadowingBuiltins,PyProtectedMember
def sum(a, axis=None, keepdims=np._NoValue, initial=np._NoValue):
    res = np.sum(a, axis=axis, keepdims=keepdims, initial=initial)
    return cast_res(res)

# Linear Algebra


def dot(a, b):
    res = np.dot(a, b)
    return cast_res(res)


def multi_dot(arrays):
    res = np.linalg.multi_dot(arrays)
    return cast_res(res)


def vdot(a, b):
    res = np.vdot(a, b)
    return cast_res(res)


def inner(a, b):
    res = np.inner(a, b)
    return cast_res(res)


def outer(a, b):
    res = np.outer(a, b)
    return cast_res(res)


def matmul(a, b):
    res = np.matmul(a, b)
    return cast_res(res)


def tensordot(a, b, axes=1):
    """
    WARNING: default value for axes is changed from 2, to 1.
    """
    res = np.tensordot(a, b, axes)
    return cast_res(res)


def kron(a, b):
    res = np.kron(a, b)
    return cast_res(res)


def trace(a, offset=0, axis1=0, axis2=1):
    res = np.trace(a, offset, axis1, axis2)
    return cast_res(res)


# Joining, splitting, and tiling arrays.


def block(arrays):
    res = np.block(arrays)
    return cast_res(res)


def concatenate(tuple_of_arrays, axis=0):
    res = np.concatenate(tuple_of_arrays, axis)
    return cast_res(res)


def stack(arrays, axis=0):
    res = np.stack(arrays, axis)
    return cast_res(res)


def column_stack(tup):
    res = np.column_stack(tup)
    return cast_res(res)


def hstack(tup):
    res = np.hstack(tup)
    return cast_res(res)


def vstack(tup):
    res = np.vstack(tup)
    return cast_res(res)


def dstack(tup):
    res = np.dstack(tup)
    return cast_res(res)


def split(array, indices_or_sections, axis=0):
    return [cast_res(v) for v in np.split(array, indices_or_sections, axis)]


def hsplit(array, indices_or_sections):
    return [cast_res(v) for v in np.hsplit(array, indices_or_sections)]


def vsplit(array, indices_or_sections):
    return [cast_res(v) for v in np.vsplit(array, indices_or_sections)]


def dsplit(array, indices_or_sections):
    return [cast_res(v) for v in np.dsplit(array, indices_or_sections)]


def array_split(array, indices_or_sections, axis=0):
    return [cast_res(v) for v in np.array_split(array, indices_or_sections, axis)]


def tile(a, reps):
    res = np.tile(a, reps)
    return cast_res(res)


def repeat(a, repeats, axis=None):
    res = np.repeat(a, repeats, axis)
    return cast_res(res)


# array creation (from data)


def diag(v, k=0):
    res = np.diag(v, k)
    return cast_res(res)


def diagflat(v, k=0):
    res = np.diagflat(v, k)
    return cast_res(res)


def tril(m, k=0):
    res = np.tril(m, k)
    return cast_res(res)


def triu(m, k=0):
    res = np.triu(m, k)
    return cast_res(res)



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt.coniclifts.constraints.set_membership.setmem import SetMembership
from sageopt.coniclifts.constraints.set_membership.product_cone import PrimalProductCone, DualProductCone
from sageopt.coniclifts.base import Variable, Expression
from sageopt.coniclifts.cones import Cone
from sageopt.coniclifts.operators import affine as aff
from sageopt.coniclifts.operators.precompiled.relent import sum_relent, elementwise_relent
from sageopt.coniclifts.operators.precompiled import affine as compiled_aff
from sageopt.coniclifts.problems.problem import Problem
import numpy as np
import warnings
from scipy.sparse import issparse
import scipy.special as special_functions

_ALLOWED_CONES_ = {'+', 'S', 'P', 'e', '0'}

_AGGRESSIVE_REDUCTION_ = True


def check_cones(K):
    if any([co.type not in _ALLOWED_CONES_ for co in K]):
        raise NotImplementedError()
    else:
        newK = [Cone(co.type, co.len) for co in K]
        return newK


class PrimalCondSageCone(SetMembership):

    def __init__(self, c, alpha, A, b, K, name, expcovers=None):
        self.name = name
        if issparse(A):
            A = A.toarray()
        self.m = alpha.shape[0]
        self.n = alpha.shape[1]
        self.lifted_n = A.shape[1]
        if self.lifted_n > self.n:
            # Then need to zero-pad alpha
            zero_block = np.zeros(shape=(alpha.shape[0], self.lifted_n - self.n))
            alpha = np.hstack((alpha, zero_block))
        self.lifted_alpha = alpha
        self.A = A
        self.b = b
        self.c = Expression(c)  # self.c is now definitely an ndarray of ScalarExpressions.
        self._variables = self.c.variables()
        self.K = check_cones(K)
        if self.m > 1:
            self.ech = ExpCoverHelper(self.lifted_alpha, self.c, self.A, self.b, self.K, expcovers)
            self.nu_vars = dict()
            self.c_vars = dict()
            self.relent_epi_vars = dict()
            self.age_vectors = dict()
            self.lambda_vars = dict()
            self._initialize_variables()
        pass

    def _initialize_variables(self):
        for i in self.ech.U_I:
            nu_len = np.count_nonzero(self.ech.expcovers[i])
            if nu_len > 0:
                var_name = 'nu^{(' + str(i) + ')}_' + self.name
                self.nu_vars[i] = Variable(shape=(nu_len,), name=var_name)
                var_name = '_relent_epi_^{(' + str(i) + ')}_' + self.name
                self.relent_epi_vars[i] = Variable(shape=(nu_len,), name=var_name)
            c_len = nu_len
            if i not in self.ech.N_I:
                c_len += 1
            var_name = 'c^{(' + str(i) + ')}_{' + self.name + '}'
            self.c_vars[i] = Variable(shape=(c_len,), name=var_name)
            var_name = 'lambda^{(' + str(i) + ')}_{' + self.name + '}'
            self.lambda_vars[i] = Variable(shape=(self.A.shape[0],), name=var_name)
        self._variables += list(self.nu_vars.values())
        self._variables += list(self.c_vars.values())
        self._variables += list(self.relent_epi_vars.values())
        self._variables += list(self.lambda_vars.values())
        pass

    def _build_aligned_age_vectors(self):
        for i in self.ech.U_I:
            ci_expr = Expression(np.zeros(self.m,))
            if i in self.ech.N_I:
                ci_expr[self.ech.expcovers[i]] = self.c_vars[i]
                ci_expr[i] = self.c[i]
            else:
                ci_expr[self.ech.expcovers[i]] = self.c_vars[i][:-1]
                ci_expr[i] = self.c_vars[i][-1]
            self.age_vectors[i] = ci_expr
        pass

    def _age_rel_ent_cone_data(self, i):
        idx_set = self.ech.expcovers[i]
        if np.any(idx_set):
            x = self.nu_vars[i]
            y = np.exp(1) * self.age_vectors[i][idx_set]
            # ^ # This line consumes a disproportionately large amount of runtime
            z = -self.age_vectors[i][i] + self.lambda_vars[i] @ self.b
            epi = self.relent_epi_vars[i]
            A_vals, A_rows, A_cols, b, K, epi = sum_relent(x, y, z, epi)
        else:
            con = self.lambda_vars[i] @ self.b <= self.age_vectors[i][i]
            con.epigraph_checked = True
            A_vals, A_rows, A_cols, b, K = con.conic_form()
        return A_vals, A_rows, A_cols, b, K

    def _age_lin_eq_cone_data(self, i):
        idx_set = self.ech.expcovers[i]
        if np.any(idx_set):
            mat1 = (self.lifted_alpha[idx_set, :] - self.lifted_alpha[i, :]).T
            mat2 = -self.A.T
            var1 = self.nu_vars[i]
            var2 = self.lambda_vars[i]
            A_vals, A_rows, A_cols = compiled_aff.mat_times_vecvar_plus_mat_times_vecvar(mat1, var1, mat2, var2)
            num_rows = mat1.shape[0] + mat2.shape[0]
        else:
            mat1 = -self.A.T.copy()
            var1 = self.lambda_vars[i]
            A_vals, A_rows, A_cols = compiled_aff.mat_times_vecvar(mat1, var1)
            num_rows = mat1.shape[0]
        b = np.zeros(num_rows, )
        K = [Cone('0', num_rows)]
        return A_vals, A_rows, A_cols, b, K

    def _age_lambda_var_domain_constraints(self, i):
        con = DualProductCone(self.lambda_vars[i], self.K)
        conic_data = con.conic_form()
        A_vals, A_rows, A_cols, b, cur_K = conic_data[0]
        return A_vals, A_rows, A_cols, b, cur_K

    def _age_violation(self, i, norm_ord, c_i, lambda_i):
        # This is "rough" only.
        lambda_viol = DualProductCone.project(lambda_i, self.K)
        if np.any(self.ech.expcovers[i]):
            idx_set = self.ech.expcovers[i]
            x_i = self.nu_vars[i].value
            x_i[x_i < 0] = 0
            y_i = np.exp(1) * c_i[idx_set]
            relent_res = np.sum(special_functions.rel_entr(x_i, y_i)) - c_i[i]  + self.b @ lambda_i # <= 0
            relent_viol = abs(max(relent_res, 0))
            eq_res = (self.lifted_alpha[idx_set, :] - self.lifted_alpha[i, :]).T @ x_i - self.A.T @ lambda_i  # == 0
            eq_res = eq_res.reshape((-1,))
            eq_viol = np.linalg.norm(eq_res, ord=norm_ord)
            total_viol = relent_viol + eq_viol + lambda_viol
            return total_viol
        else:
            c_i = float(self.c_vars[i].value)
            residual = c_i - self.b @ lambda_i  # >= 0
            relent_viol = abs(min(0, residual))
            total_viol = relent_viol + lambda_viol
            return total_viol

    def _age_vectors_sum_to_c(self):
        nonconst_locs = np.ones(self.m, dtype=bool)
        nonconst_locs[self.ech.N_I] = False
        aux_c_vars = list(self.age_vectors.values())
        aux_c_vars = aff.vstack(aux_c_vars).T
        aux_c_vars = aux_c_vars[nonconst_locs, :]
        main_c_var = self.c[nonconst_locs]
        # add constraint that the columns of all_cs sum to self.c[need_constr]
        # TODO: make "sums of vectors equaling another vector" a precompiled atom.
        A_rows, A_cols, A_vals = [], [], []
        num_cons = np.count_nonzero(nonconst_locs)
        b = np.zeros(num_cons,)
        K = [Cone('0', num_cons)]
        for i in range(num_cons):
            # update cols and data to reflect addition of elements in ith row of aux_c_vars
            svs = aux_c_vars[i, :].scalar_variables()
            A_cols += [sv.id for sv in svs]
            A_vals += [1] * len(svs)
            # update cols and data to reflect addition of elements from ith element of main_c_var
            #   ith element of main_c_var is a ScalarExpression!
            id2co = [(a.id, co) for a, co in main_c_var[i].atoms_to_coeffs.items()]
            A_cols += [aid for aid, _ in id2co]
            A_vals += [-co for _, co in id2co]  # we are subtracting, after all.
            # update rows with appropriate number of "i"s.
            A_rows += [i] * (len(svs) + len(id2co))
            # update b
            b[i] -= main_c_var[i].offset
        return A_vals, np.array(A_rows), A_cols, b, K

    def variables(self):
        return self._variables

    def conic_form(self):
        if self.m > 1:
            # Lift c_vars and nu_vars into Expressions of length self.m
            self._build_aligned_age_vectors()
            cone_data = []
            # age cones
            for i in self.ech.U_I:
                cone_data.append(self._age_rel_ent_cone_data(i))
                cone_data.append(self._age_lin_eq_cone_data(i))
                cone_data.append(self._age_lambda_var_domain_constraints(i))
            # Vectors sum to s.c
            cone_data.append(self._age_vectors_sum_to_c())
            return cone_data
        else:
            con = self.c >= 0
            con.epigraph_checked = True
            A_vals, A_rows, A_cols, b, K = con.conic_form()
            cone_data = [(A_vals, A_rows, A_cols, b, K)]
            return cone_data

    def violation(self, norm_ord=np.inf, rough=False):
        c = self.c.value
        if self.m > 1:
            if not rough and c in self:
                return 0
            # compute violation for "AGE vectors sum to c"
            #   Although, we can use the fact that the SAGE cone contains R^m_++.
            #   and so only compute violation for "AGE vectors sum to <= c"
            age_vectors = {i: v.value for i, v in self.age_vectors.items()}
            lambda_vectors = {i: v.value for i, v in self.lambda_vars.items()}
            sum_age_vectors = sum(age_vectors.values())
            residual = c - sum_age_vectors  # want >= 0
            residual[residual < 0] = 0
            sum_to_c_viol = np.linalg.norm(residual, ord=norm_ord)
            # compute violations for each AGE cone
            age_viols = np.zeros(shape=(len(self.ech.U_I,)))
            for idx, i in enumerate(self.ech.U_I):
                age_viols[idx] = self._age_violation(i, norm_ord, age_vectors[i], lambda_vectors[i])
            # add the max "AGE violation" to the violation for "AGE vectors sum to c".
            total_viol = sum_to_c_viol + np.max(age_viols)
            return total_viol
        else:
            residual = c.reshape((-1,))  # >= 0
            residual[residual >= 0] = 0
            return np.linalg.norm(c, ord=norm_ord)
        pass

    def __contains__(self, item):
        from sageopt.coniclifts import clear_variable_indices, MIN as CL_MIN, SOLVED as CL_SOLVED
        item = Expression(item).ravel()
        name = self.name + ' check membership'
        con = [PrimalCondSageCone(item, self.lifted_alpha, self.A, self.b, self.K, name)]
        prob = Problem(CL_MIN, Expression([0]), con)
        prob.solve(verbose=False)
        clear_variable_indices()
        return prob.status == CL_SOLVED and abs(prob.value) < 1e-7


class DualCondSageCone(SetMembership):

    def __init__(self, v, alpha, A, b, K, name, c=None, expcovers=None):
        """
        Aggregates constraints on "v" so that "v" can be viewed as a dual variable
        to a constraint of the form "c \in C_{SAGE}(alpha, A, b, K)".
        """
        if issparse(A):
            A = A.toarray()
        self.A = A
        self.b = b
        self.K = check_cones(K)
        self.n = alpha.shape[1]
        self.m = alpha.shape[0]
        self.lifted_n = A.shape[1]
        if self.lifted_n > self.n:
            zero_block = np.zeros(shape=(alpha.shape[0], self.lifted_n - self.n))
            alpha = np.hstack((alpha, zero_block))
        self.lifted_alpha = alpha
        self.v = v
        self.name = name
        self._variables = self.v.variables()
        if self.m > 1:
            if c is None:
                self.c = None
            else:
                self.c = Expression(c)
            self.ech = ExpCoverHelper(self.lifted_alpha, self.c, self.A, self.b, self.K, expcovers)
            self.lifted_mu_vars = dict()
            self.mu_vars = dict()
            self.relent_epi_vars = dict()
            self._initialize_variables()
        pass

    def _initialize_variables(self):
        for i in self.ech.U_I:
            var_name = 'mu[' + str(i) + ']_{' + self.name + '}'
            self.lifted_mu_vars[i] = Variable(shape=(self.lifted_n,), name=var_name)
            self._variables.append(self.lifted_mu_vars[i])
            self.mu_vars[i] = self.lifted_mu_vars[i][:self.n]

            num_cover = np.count_nonzero(self.ech.expcovers[i])
            if num_cover > 0:
                var_name = '_relent_epi_[' + str(i) + ']_{' + self.name + '}'
                epi = Variable(shape=(num_cover,), name=var_name)
                self.relent_epi_vars[i] = epi
                self._variables.append(epi)
        pass

    def _dual_age_cone_data(self, i):
        cone_data = []
        selector = self.ech.expcovers[i]
        len_sel = np.count_nonzero(selector)
        if len_sel > 0:
            #
            # relative entropy constraints
            #
            expr1 = np.tile(self.v[i], len_sel).view(Expression)
            epi = self.relent_epi_vars[i]
            A_vals, A_rows, A_cols, b, K, epi = elementwise_relent(expr1, self.v[selector], epi)
            cone_data.append((A_vals, A_rows, A_cols, b, K))
            #
            # Linear inequalities
            #
            matrix = self.lifted_alpha[selector, :] - self.lifted_alpha[i, :]
            vecvar = self.lifted_mu_vars[i]
            A_vals, A_rows, A_cols = compiled_aff.mat_times_vecvar_minus_vecvar(-matrix, vecvar, epi)
            num_rows = matrix.shape[0]
            b = np.zeros(num_rows)
            K = [Cone('+', num_rows)]
            A_rows = np.array(A_rows)
            cone_data.append((A_vals, A_rows, A_cols, b, K))
        #
        # the additional constraints, for the generalized AGE dual cone
        #
        mat = self.A
        vecvar = self.lifted_mu_vars[i]
        vec = self.b
        singlevar = self.v[i]
        A_vals, A_rows, A_cols = compiled_aff.mat_times_vecvar_plus_vec_times_singlevar(mat, vecvar, vec, singlevar)
        b = np.zeros(self.A.shape[0], )
        cur_K = [Cone(co.type, co.len) for co in self.K]
        cone_data.append((A_vals, A_rows, A_cols, b, cur_K))
        return cone_data

    def _dual_age_cone_violation(self, i, norm_ord, rough, v):
        from sageopt.coniclifts import clear_variable_indices, MIN as CL_MIN, SOLVED as CL_SOLVED
        selector = self.ech.expcovers[i]
        len_sel = np.count_nonzero(selector)
        # TODO: fix how this function handles len_sel == 0.
        #   The current behavior isn't valid when {x: A @ x + b \in K} is the empty set.
        if len_sel > 0:
            expr1 = np.tile(v[i], len_sel).ravel()
            expr2 = v[selector].ravel()
            lowerbounds = special_functions.rel_entr(expr1, expr2)
            mat = -(self.lifted_alpha[selector, :] - self.lifted_alpha[i, :])
            mu_i = self.lifted_mu_vars[i].value
            # compute rough violation for this dual AGE cone
            residual = mat @ mu_i - lowerbounds
            residual[residual >= 0] = 0
            relent_viol = np.linalg.norm(residual, ord=norm_ord)
            AbK_val = self.A @ mu_i + v[i] * self.b
            AbK_viol = PrimalProductCone.project(AbK_val, self.K)
            curr_viol = relent_viol + AbK_viol
            # as applicable, solve an optimization problem to compute the violation.
            if curr_viol > 0 and not rough:
                temp_var = Variable(shape=(mat.shape[1],), name='temp_var')
                cons = [mat @ temp_var >= lowerbounds,
                        PrimalProductCone(self.A @ temp_var + v[i] * self.b, self.K)]
                prob = Problem(CL_MIN, Expression([0]), cons)
                status, value = prob.solve(verbose=False)
                clear_variable_indices()
                if status == CL_SOLVED and abs(value) < 1e-7:
                    curr_viol = 0
            return curr_viol
        else:
            return 0

    def variables(self):
        return self._variables

    def conic_form(self):
        if self.m > 1:
            nontrivial_I = list(set(self.ech.U_I + self.ech.P_I))
            con = self.v[nontrivial_I] >= 0
            con.epigraph_checked = True
            A_vals, A_rows, A_cols, b, K = con.conic_form()
            cone_data = [(A_vals, A_rows, A_cols, b, K)]
            for i in self.ech.U_I:
                curr_age = self._dual_age_cone_data(i)
                cone_data += curr_age
            return cone_data
        else:
            con = self.v >= 0
            con.epigraph_checked = True
            A_vals, A_rows, A_cols, b, K = con.conic_form()
            cone_data = [(A_vals, A_rows, A_cols, b, K)]
            return cone_data

    def violation(self, norm_ord=np.inf, rough=False):
        v = self.v.value
        viols = []
        for i in self.ech.U_I:
            curr_viol = self._dual_age_cone_violation(i, norm_ord, rough, v)
            viols.append(curr_viol)
        viol = max(viols)
        return viol


class ExpCoverHelper(object):

    def __init__(self, alpha, c, A, b, K, expcovers=None):
        if c is not None and not isinstance(c, Expression):
            raise RuntimeError()
        self.alpha = alpha
        self.A = A
        self.b = b
        self.K = K
        self.m = alpha.shape[0]
        self.c = c
        if self.c is not None:
            self.U_I = [i for i, c_i in enumerate(self.c) if (not c_i.is_constant()) or (c_i.offset < 0)]
            # ^ indices of not-necessarily-positive sign; i \in U_I must get an AGE cone.
            # this AGE cone might not be used in the final solution to the associated
            # optimization problem.
            self.N_I = [i for i, c_i in enumerate(self.c) if (c_i.is_constant()) and (c_i.offset < 0)]
            # ^ indices of definitively-negative sign. if j \in N_I, then there is only one AGE
            # cone (indexed by i) with c^{(i)}_j != 0, and that's j == i. These AGE cones will
            # be used in any solution to the associated optimization problem, and we can be
            # certain that c^{(i)}_i == c_i.
            self.P_I = [i for i, c_i in enumerate(self.c) if (c_i.is_constant()) and (c_i.offset > 0)]
            # ^ indices of positive sign.
            # Together, union(self.P_I, self.U_I) comprise all indices "i" where c[i] is not identically zero.
        else:
            self.U_I = [i for i in range(self.m)]
            self.N_I = []
            self.P_I = []
        if isinstance(expcovers, dict):
            self._verify_exp_covers(expcovers)
        elif expcovers is None:
            expcovers = self._default_exp_covers()
        else:
            raise RuntimeError('Argument "expcovers" must be a dict.')
        self.expcovers = expcovers

    def _verify_exp_covers(self, expcovers):
        for i in self.U_I:
            if i not in expcovers:
                raise RuntimeError('Required key missing from "expcovers".')
            if (not isinstance(expcovers[i], np.ndarray)) or (expcovers[i].dtype != bool):
                raise RuntimeError('A value in "expcovers" was the wrong datatype.')
            if expcovers[i][i]:
                warnings.warn('Nonsensical value in "expcovers"; correcting.')
                expcovers[i][i] = False
        pass

    def _default_exp_covers(self):
        expcovers = dict()
        for i in self.U_I:
            cov = np.ones(shape=(self.m,), dtype=bool)
            cov[self.N_I] = False
            cov[i] = False
            expcovers[i] = cov
        if _AGGRESSIVE_REDUCTION_:
            row_sums = np.sum(self.alpha, 1)
            if np.all(self.alpha >= 0) and np.min(row_sums) == 0:
                # Then apply the reduction.
                zero_loc = np.nonzero(row_sums == 0)[0][0]
                for i in self.U_I:
                    if i == zero_loc:
                        continue
                    curr_cover = expcovers[i]
                    curr_row = self.alpha[i, :]
                    for j in range(self.m):
                        if curr_cover[j] and j != zero_loc and curr_row @ self.alpha[j, :] == 0:
                            curr_cover[j] = False
        return expcovers



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import sageopt.coniclifts as cl
import numpy as np
from collections import defaultdict


__NUMERIC_TYPES__ = (int, float, np.int_, np.float_)

__EXPONENT_VECTOR_DECIMAL_POINTS__ = 7


def standard_sig_monomials(n):
    """
    Parameters
    ----------
    n : int
        The dimension of the space over which the constituent Signomials will be defined.

    Returns
    -------
    y : NumPy ndarray
        An array  of length ``n``, with ``y[i]`` as a Signomial with one term,
        corresponding to the (``i+1``)-th standard basis vector in ``n`` dimensional real space.

    Example
    -------
    This function is useful for constructing signomials in an algebraic form. ::

        y = standard_sig_monomials(3)
        f = (y[0] ** 1.5) * (y[2] ** -0.6) - y[1] * y[2]

    """
    y = np.empty(shape=(n,), dtype=object)
    for i in range(n):
        ei = np.zeros(shape=(1, n))
        ei[0, i] = 1
        y[i] = Signomial(ei, np.array([1]))
    return y


class Signomial(object):
    """
    This class provides a symbolic representation for linear combinations of exponentials,
    composed with linear functions. These functions look like the following.

    .. math::

       x \mapsto \sum_{i=1}^m c_i \exp({\\alpha}_i \cdot x)

    The constructor for this class can be called in two different ways. The arguments
    to the constructor have names which reflect the different possibilities.

    Parameters
    ----------

    alpha_maybe_c : dict or NumPy ndarray

         If ``alpha_maybe_c`` is a dict, then it must be a dictionary from tuples-of-numbers to
         scalars. The keys will be converted to rows of a matrix which we call ``alpha``, and
         the values will be assembled into a vector which we call ``c``.

         If ``alpha_maybe_c`` is a NumPy ndarray, then the argument ``c`` must also be an ndarray,
         and ``c.size`` must equal ``alpha_maybe_c.shape[0]``.

    c : None or NumPy ndarray

        This value is only used when ``alpha_maybe_c`` is a NumPy ndarray. If that is the case, then
        this Signomial will represent the function ``lambda x: c @ np.exp(alpha_maybe_c @ x)``.

    Examples
    --------

    There are two ways to call the Signomial constructor.

    The first way is to specify a dictionary from tuples to scalars. The tuples are interpreted as linear
    functionals appearing in the exponential terms, and the scalars are the corresponding coefficients.::

        alpha_and_c = {(1,): 2}
        f = Signomial(alpha_and_c)
        print(f(1))  # equal to 2 * np.exp(1).
        print(f(0))  # equal to 2.

    The second way is to specify two arguments. In this case the first argument is a NumPy array
    where the rows represent linear functionals, and the second argument is a vector of corresponding
    coefficients.::

        alpha = np.array([[1, 0], [0, 1], [1, 1]])
        c = np.array([1, 2, 3])
        f = Signomial(alpha, c)
        x = np.random.randn(2)
        print(f(x) - c @ np.exp(alpha @ x))  # zero, up to rounding errors.

    Attributes
    ----------

    n : int
        The dimension of the space over which this Signomial is defined. The number of columns in ``alpha``,
        and the length of tuples appearing in the dictionary ``alpha_c``.

    m : int
        The number of terms needed to express this Signomial in a basis of monomial functions
        ``lambda x: exp(a @ x)`` for row vectors ``a``. The signomial is presumed to include a constant term.

    alpha : NumPy ndarray
        Has shape ``(m, n)``. Each row specifies a vector appearing in an exponential function which
        defines this Signomial. The rows are ordered for consistency with the property ``c``.

    c : NumPy ndarray
        Has shape ``(m,)``. The scalar ``c[i]`` is this Signomial's coefficient on the basis function
        ``lambda x: exp(alpha[i, :] @ x)``. It is possible to have ``c.dtype == object``.

    alpha_c : defaultdict
        The keys of ``alpha_c`` are tuples of length ``n``, containing real numeric types (e.g int, float).
        These tuples define linear functions. This Signomial could be evaluated by the code snippet
        ``lambda x: np.sum([ alpha_c[a] * np.exp(a @ x) for a in alpha_c])``. The default value for this
        dictionary is zero.

    Notes
    -----

    Operator overloading.

        The Signomial class implements ``+``, ``-``, ``*``, and ``**`` between pairs of Signomials,
        and pairs involving one Signomial and one numeric type.

        The Signomial class also implements ``s1 / s2`` where ``s2`` is a numeric type or Signomial,
        but if ``s2`` is a Signomial, then its coefficient vector ``s2.c`` can only contain one nonzero entry.

        You can also use ``+``, ``-``, and ``*`` for pairs involving one Signomial and one non-Signomial.
        If ``s3`` is the result of such an operation, then ``s3.c`` will be a NumPy array with ``s3.dtype == object``.

    Function evaluations.

        Signomial objects are callable. If ``s`` is a Signomial object and ``x`` is a numpy array of length ``s.n``,
        then ``s(x)`` computes the Signomial object as though it were any other elementary Python function.

        Signomial objects provide functions to compute gradients (equivalently, Jacobians) and Hessians.
        These methods operate by caching and evaluating symbolic representations of the relevant partial derivatives.

    Internal representations.

        Both ``self.alpha_c`` and ``(self.alpha, self.c)`` completely specify a Signomial object.

        ``alpha_c`` is the dictionary which is taken to *define* this Signomial as a mathematical object.

        However, it is useful to have rapid access to the matrix of linear functionals ``alpha``, or the coefficient
        vector ``c`` as numpy arrays. So these fields are also maintained explicitly.

        If a user modifies the fields ``alpha`` or ``c`` directly, there may be an inconsistency between these
        fields, and the dictionary ``alpha_c``. Therefore the fields ``alpha`` and ``c`` should not be modified
        without taking great care to ensure consistency with the signomial's dictionary representation.
    """

    def __init__(self, alpha_maybe_c, c=None):
        # noinspection PyArgumentList
        if c is None:
            # noinspection PyArgumentList
            self.alpha_c = defaultdict(int, alpha_maybe_c)
        else:
            alpha = np.round(alpha_maybe_c, decimals=__EXPONENT_VECTOR_DECIMAL_POINTS__)
            alpha = alpha.tolist()
            if len(alpha) != c.size:
                raise RuntimeError('alpha and c specify different numbers of terms')
            self.alpha_c = defaultdict(int)
            for j in range(c.size):
                self.alpha_c[tuple(alpha[j])] += c[j]
        self.n = len(list(self.alpha_c.items())[0][0])
        self.alpha_c[(0,) * self.n] += 0  # ensures that there's a constant term.
        self.m = len(self.alpha_c)
        self.alpha = None
        self.c = None
        self._grad = None
        self._hess = None
        self._update_alpha_c_arrays()
        pass

    def _cache_grad(self):
        if self._grad is None:
            g = np.empty(shape=(self.n,), dtype=object)
            for i in range(self.n):
                g[i] = self.partial(i)
            self._grad = g
        pass

    def _cache_hess(self):
        if self._hess is None:
            H = np.empty(shape=(self.n, self.n), dtype=object)
            for i in range(self.n):
                ith_partial = self.partial(i)
                for j in range(i+1):
                    curr_partial = ith_partial.partial(j)
                    H[i, j] = curr_partial
                    H[j, i] = curr_partial
            self._hess = H
        pass

    @property
    def grad(self):
        """
        A numpy ndarray of shape ``(n,)`` whose entries are Signomials. For a numpy ndarray ``x``, ``grad[i](x)``
        is the partial derivative of this Signomial with respect to coordinate ``i``, evaluated at ``x``.
        This array is constructed only when necessary, and is cached upon construction.
        """
        self._cache_grad()
        return self._grad


    @property
    def hess(self):
        """
        A numpy ndarray of shape ``(n, n)``, whose entries are Signomials. For a numpy ndarray ``x``,
        ``hess[i,j](x)`` is the (i,j)-th partial derivative of this Signomial, evaluated at ``x``.
        This array is constructed only when necessary, and is cached upon construction.
        """
        self._cache_hess()
        return self._hess

    def query_coeff(self, a):
        """
        Returns the coefficient of the basis function ``lambda x: np.exp(a @ x)`` for this Signomial.
        """
        tup = tuple(np.round(a, decimals=__EXPONENT_VECTOR_DECIMAL_POINTS__))
        if tup in self.alpha_c:
            return self.alpha_c[tup]
        else:
            return 0

    def constant_location(self):
        """
        Return the index ``i`` so that ``alpha[i, :]`` is the zero vector.
        """
        return np.where((self.alpha == 0).all(axis=1))[0][0]

    def _update_alpha_c_arrays(self):
        """
        Call this function whenever the dictionary representation of this Signomial object has been updated.
        """
        alpha = []
        c = []
        for k, v in self.alpha_c.items():
            alpha.append(k)
            c.append(v)
        self.alpha = np.array(alpha)
        self.c = np.array(c)
        if self.c.dtype is object:
            self.c = cl.Expression(self.c)

    def __add__(self, other):
        if not isinstance(other, Signomial):
            tup = (0,) * self.n
            d = {tup: other}
            other = Signomial(d)
        # noinspection PyArgumentList
        d = defaultdict(int, self.alpha_c)
        for k, v in other.alpha_c.items():
            d[k] += v
        res = Signomial(d)
        res.remove_terms_with_zero_as_coefficient()
        return res

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        if not isinstance(other, Signomial):
            tup = (0,) * self.n
            d = {tup: other}
            other = Signomial(d)
        d = defaultdict(int)
        alpha1, c1 = self.alpha, self.c
        alpha2, c2 = other.alpha, other.c
        for i1, v1 in enumerate(alpha1):
            for i2, v2 in enumerate(alpha2):
                v3 = np.round(v1 + v2, decimals=__EXPONENT_VECTOR_DECIMAL_POINTS__)
                d[tuple(v3)] += c1[i1] * c2[i2]
        res = Signomial(d)
        res.remove_terms_with_zero_as_coefficient()
        return res

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        other_inv = other ** -1
        return self.__mul__(other_inv)

    def __rtruediv__(self, other):
        self_inv = self ** -1
        return self_inv * other

    def __sub__(self, other):
        # noinspection PyTypeChecker
        return self.__add__(-1 * other)

    def __rsub__(self, other):
        # noinspection PyTypeChecker
        return other + (-1) * self

    def __pow__(self, power, modulo=None):
        if self.c.dtype not in __NUMERIC_TYPES__:
            raise RuntimeError('Cannot exponentiate signomials with symbolic coefficients.')
        if power % 1 == 0 and power >= 0:
            power = int(power)
            if power == 0:
                # noinspection PyTypeChecker
                return Signomial({(0,) * self.n: 1})
            else:
                s = Signomial(self.alpha_c)
                for t in range(power - 1):
                    s = s * self
                return s
        else:
            d = dict((k, v) for (k, v) in self.alpha_c.items() if v != 0)
            if len(d) != 1:
                raise ValueError('Only signomials with exactly one term can be raised to power %s.')
            v = list(d.values())[0]
            if v < 0 and not power % 1 == 0:
                raise ValueError('Cannot compute non-integer power %s of coefficient %s', power, v)
            alpha_tup = tuple(power * ai for ai in list(d.keys())[0])
            c = float(v) ** power
            s = Signomial(alpha_maybe_c={alpha_tup: c})
            return s

    def __neg__(self):
        # noinspection PyTypeChecker
        return self.__mul__(-1)

    def __call__(self, x):
        """
        Evaluates the mathematical function specified by the current Signomial object.

        :param x: either a scalar (if self.n == 1), or a numpy n-d array with len(x.shape) <= 2
        and x.shape[0] == self.n.
        :return:  If x is a scalar or an n-d array of shape (self.n,), then "val" is a numeric
        type equal to the signomial evaluated at x. If instead x is of shape (self.n, k) for
        some positive integer k, then "val" is a numpy n-d array of shape (k,), with val[i]
        equal to the current signomial evaluated on the i^th column of x.

        This function's behavior is undefined when x is not a scalar and has len(x.shape) > 2.
        """
        if np.isscalar(x) or (isinstance(x, np.ndarray) and x.size == 1 and x.ndim == 0):
            x = np.array([np.asscalar(np.array(x))])  # coerce into a 1d array of shape (1,).
        if not x.shape[0] == self.n:
            msg = 'Domain is R^' + str(self.n) + 'but x is in R^' + str(x.shape[0])
            raise ValueError(msg)
        if x.ndim > 2:
            msg = 'Signomials cannot be called on ndarrays with more than 2 dimensions.'
            raise ValueError(msg)
        x = x.astype(np.float128)
        exponents = np.dot(self.alpha.astype(np.float128), x)
        linear_vars = np.exp(exponents).astype(np.float128)
        val = np.dot(self.c, linear_vars)
        return val

    def __hash__(self):
        return hash(frozenset(self.alpha_c.items()))

    def __eq__(self, other):
        if not isinstance(other, Signomial):
            return False
        if self.m != other.m:
            return False
        for k in self.alpha_c:
            v = self.alpha_c[k]
            other_v = other.alpha_c[k]
            if not cl.Expression.are_equivalent(other_v, v, rtol=1e-8):
                return False
        return True

    def remove_terms_with_zero_as_coefficient(self):
        """
        Update ``alpha``, ``c``, and ``alpha_c`` to remove nonconstant terms where ``c[i] == 0``.
        """
        d = dict()
        for (k, v) in self.alpha_c.items():
            if (not isinstance(v, __NUMERIC_TYPES__)) or v != 0:
                d[k] = v
        # noinspection PyArgumentList
        self.alpha_c = defaultdict(int, d)
        tup = (0,) * self.n
        self.alpha_c[tup] += 0
        self.m = len(self.alpha_c)
        self._update_alpha_c_arrays()
        pass

    def partial(self, i):
        """
        Compute the symbolic partial derivative of this signomial, at coordinate ``i``.

        Parameters
        ----------
        i : int
            ``i`` must be an integer from 0 to ``self.n - 1``.

        Returns
        -------
        p : Signomial
            The function obtained by differentiating this signomial with respect to its i-th argument.
        """
        if i < 0 or i >= self.n:
            raise RuntimeError('This Signomial does not have an input at index ' + str(i) + '.')
        d = defaultdict(int)
        for j in range(self.m):
            c = self.c[j] * self.alpha[j, i]
            if c != 0:
                vec = self.alpha[j, :].copy()
                vec[i] -= 1
                d[tuple(vec.tolist())] = c
        d[self.n * (0,)] += 0
        p = Signomial(d)
        return p

    def grad_val(self, x):
        """
        Return the gradient of this Signomial (as a NumPy ndarray) at the point ``x``.
        """
        _ = self.grad  # construct the function handles.
        g = np.zeros(self.n)
        for i in range(self.n):
            g[i] = self._grad[i](x)
        return g

    def hess_val(self, x):
        """
        Return the Hessian of this Signomial (as a NumPy ndarray) at the point ``x``.
        """
        if self._hess is None:
            _ = self.hess  # ignore the return value
        H = np.zeros(shape=(self.n, self.n))
        for i in range(self.n):
            for j in range(i+1):
                val = self._hess[i, j](x)
                H[i, j] = val
                H[j, i] = val
        return H

    def as_polynomial(self):
        """
        This function is only applicable if ``alpha`` is a matrix of nonnegative integers.

        Returns
        -------
        f : Polynomial
            For every elementwise vector ``x``, we have ``self(x) == f(np.exp(x))``.
        """
        from sageopt.symbolic.polynomials import Polynomial
        f = Polynomial(self.alpha, self.c)
        return f
